@@ -20,25 +20,28 @@
  * ----------------------------------------------------------------------- */
 
 #include <dirent.h>
-#include <libgen.h>
+#include <fcntl.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/mount.h>
-#include <sys/vfs.h>
 #include <sched.h>
+#include <pwd.h>
+#include <grp.h>
 
-#define INCLUDE_PENDING_FUNCTIONS
 #include "automount.h"
 
-/* Attribute to create detached thread */
-extern pthread_attr_t th_attr_detached;
+extern pthread_attr_t thread_attr;
+
+static pthread_mutex_t ma_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t ea_mutex = PTHREAD_MUTEX_INITIALIZER;
 
 static int unlink_mount_tree(struct autofs_point *ap, struct mnt_list *mnts)
 {
@@ -62,6 +65,7 @@ static int unlink_mount_tree(struct autofs_point *ap, struct mnt_list *mnts)
 		else
 			rv = umount2(this->path, MNT_DETACH);
 		if (rv == -1) {
+			ret = 0;
 			debug(ap->logopt,
 			      "can't unlink %s from mount tree", this->path);
 
@@ -73,7 +77,6 @@ static int unlink_mount_tree(struct autofs_point *ap, struct mnt_list *mnts)
 
 			case ENOENT:
 			case EFAULT:
-				ret = 0;
 				warn(ap->logopt, "bad path for mount");
 				break;
 			}
@@ -83,116 +86,105 @@ static int unlink_mount_tree(struct autofs_point *ap, struct mnt_list *mnts)
 	return ret;
 }
 
-static int do_mount_autofs_indirect(struct autofs_point *ap, const char *root)
+static int do_mount_autofs_indirect(struct autofs_point *ap)
 {
-	const char *str_indirect = mount_type_str(t_indirect);
-	struct ioctl_ops *ops = get_ioctl_ops();
-	time_t timeout = get_exp_timeout(ap, ap->entry->maps);
+	time_t timeout = ap->exp_timeout;
 	char *options = NULL;
-	const char *hosts_map_name = "-hosts";
-	const char *map_name = hosts_map_name;
-	const char *type;
+	const char *type, *map_name = NULL;
 	struct stat st;
 	struct mnt_list *mnts;
-	int ret;
-	int err;
+	int cl_flags, ret;
 
-	/* If the map is being shared the exp_timeout can't be inherited
-	 * from the map source since it may be different so the autofs
-	 * point exp_runfreq must have already been set.
-	 */
-	if (ap->entry->maps->ref <= 1)
-		ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
-
-	if (ops->version && !do_force_unlink) {
-		ap->flags |= MOUNT_FLAG_REMOUNT;
-		ret = try_remount(ap, NULL, t_indirect);
-		ap->flags &= ~MOUNT_FLAG_REMOUNT;
-		if (ret == 1)
-			return 0;
-		if (ret == 0)
-			return -1;
-	} else {
-		mnts = get_mnt_list(_PROC_MOUNTS, ap->path, 1);
-		if (mnts) {
-			ret = unlink_mount_tree(ap, mnts);
-			free_mnt_list(mnts);
-			if (!ret) {
-				error(ap->logopt,
-				      "already mounted as other than autofs "
-				      "or failed to unlink entry in tree");
-				goto out_err;
-			}
+	mnts = get_mnt_list(_PROC_MOUNTS, ap->path, 1);
+	if (mnts) {
+		ret = unlink_mount_tree(ap, mnts);
+		free_mnt_list(mnts);
+		if (!ret) {
+			debug(ap->logopt,
+			      "already mounted as other than autofs "
+			      "or failed to unlink entry in tree");
+			goto out_err;
 		}
 	}
 
-	options = make_options_string(ap->path, ap->kpipefd, str_indirect);
-	if (!options) {
-		error(ap->logopt, "options string error");
+	options = make_options_string(ap->path, ap->kpipefd, NULL);
+	if (!options)
 		goto out_err;
-	}
 
 	/* In case the directory doesn't exist, try to mkdir it */
-	if (mkdir_path(root, 0555) < 0) {
+	if (mkdir_path(ap->path, 0555) < 0) {
 		if (errno != EEXIST && errno != EROFS) {
 			crit(ap->logopt,
 			     "failed to create autofs directory %s",
-			     root);
+			     ap->path);
 			goto out_err;
 		}
 		/* If we recieve an error, and it's EEXIST or EROFS we know
 		   the directory was not created. */
-		ap->flags &= ~MOUNT_FLAG_DIR_CREATED;
+		ap->dir_created = 0;
 	} else {
 		/* No errors so the directory was successfully created */
-		ap->flags |= MOUNT_FLAG_DIR_CREATED;
+		ap->dir_created = 1;
 	}
 
 	type = ap->entry->maps->type;
-	if (!type || strcmp(ap->entry->maps->type, "hosts"))
+	if (type && !strcmp(ap->entry->maps->type, "hosts")) {
+		char *tmp = alloca(7);
+		if (tmp) {
+			strcpy(tmp, "-hosts");
+			map_name = (const char *) tmp;
+		}
+	} else
 		map_name = ap->entry->maps->argv[0];
 
-	ret = mount(map_name, root, "autofs", MS_MGC_VAL, options);
+	ret = mount(map_name, ap->path, "autofs", MS_MGC_VAL, options);
 	if (ret) {
-		crit(ap->logopt,
-		     "failed to mount autofs path %s at %s", ap->path, root);
+		crit(ap->logopt, "failed to mount autofs path %s", ap->path);
 		goto out_rmdir;
 	}
 
 	free(options);
+
 	options = NULL;
 
-	ret = stat(root, &st);
-	if (ret == -1) {
-		crit(ap->logopt,
-		     "failed to stat mount for autofs path %s", ap->path);
-		goto out_umount;
-	}
-
-	if (ap->mode && (err = chmod(root, ap->mode)))
-		warn(ap->logopt, "failed to change mode of %s", ap->path);
-
-	if (ops->open(ap->logopt, &ap->ioctlfd, st.st_dev, root)) {
+	/* Root directory for ioctl()'s */
+	ap->ioctlfd = open(ap->path, O_RDONLY);
+	if (ap->ioctlfd < 0) {
 		crit(ap->logopt,
 		     "failed to create ioctl fd for autofs path %s", ap->path);
 		goto out_umount;
 	}
 
-	ap->dev = st.st_dev;	/* Device number for mount point checks */
+	if ((cl_flags = fcntl(ap->ioctlfd, F_GETFD, 0)) != -1) {
+		cl_flags |= FD_CLOEXEC;
+		fcntl(ap->ioctlfd, F_SETFD, cl_flags);
+	}
 
-	ops->timeout(ap->logopt, ap->ioctlfd, timeout);
-	if (ap->logopt & LOGOPT_DEBUG)
-		notify_mount_result(ap, root, timeout, str_indirect);
+	ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
+
+	ioctl(ap->ioctlfd, AUTOFS_IOC_SETTIMEOUT, &timeout);
+
+	if (ap->exp_timeout)
+		info(ap->logopt,
+		    "mounted indirect mount on %s "
+		    "with timeout %u, freq %u seconds", ap->path,
+	 	    (unsigned int) ap->exp_timeout,
+		    (unsigned int) ap->exp_runfreq);
 	else
-		notify_mount_result(ap, ap->path, timeout, str_indirect);
+		info(ap->logopt,
+		    "mounted indirect mount on %s with timeouts disabled",
+		    ap->path);
+
+	fstat(ap->ioctlfd, &st);
+	ap->dev = st.st_dev;	/* Device number for mount point checks */
 
 	return 0;
 
 out_umount:
-	umount(root);
+	umount(ap->path);
 out_rmdir:
-	if (ap->flags & MOUNT_FLAG_DIR_CREATED)
-		rmdir(root);
+	if (ap->dir_created)
+		rmdir_path(ap, ap->path, ap->dev);
 out_err:
 	if (options)
 		free(options);
@@ -204,9 +196,9 @@ out_err:
 	return -1;
 }
 
-int mount_autofs_indirect(struct autofs_point *ap, const char *root)
+int mount_autofs_indirect(struct autofs_point *ap)
 {
-	time_t now = monotonic_time(NULL);
+	time_t now = time(NULL);
 	int status;
 	int map;
 
@@ -218,11 +210,11 @@ int mount_autofs_indirect(struct autofs_point *ap, const char *root)
 		return -1;
 	}
 
-	status = do_mount_autofs_indirect(ap, root);
+	status = do_mount_autofs_indirect(ap);
 	if (status < 0)
 		return -1;
 
-	map = lookup_ghost(ap, root);
+	map = lookup_ghost(ap);
 	if (map & LKP_FAIL) {
 		if (map & LKP_DIRECT) {
 			error(ap->logopt,
@@ -236,13 +228,16 @@ int mount_autofs_indirect(struct autofs_point *ap, const char *root)
 	}
 
 	if (map & LKP_NOTSUP)
-		ap->flags &= ~MOUNT_FLAG_GHOST;
+		ap->ghost = 0;
 
 	return 0;
 }
 
-void close_mount_fds(struct autofs_point *ap)
+int umount_autofs_indirect(struct autofs_point *ap)
 {
+	char buf[MAX_ERR_BUF];
+	int ret, rv, retries;
+
 	/*
 	 * Since submounts look after themselves the parent never knows
 	 * it needs to close the ioctlfd for offset mounts so we have
@@ -252,6 +247,20 @@ void close_mount_fds(struct autofs_point *ap)
 	if (ap->submount)
 		lookup_source_close_ioctlfd(ap->parent, ap->path);
 
+	/* If we are trying to shutdown make sure we can umount */
+	rv = ioctl(ap->ioctlfd, AUTOFS_IOC_ASKUMOUNT, &ret);
+	if (rv == -1) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr("ioctl failed: %s", estr);
+		return 1;
+	} else if (!ret) {
+		error(ap->logopt, "ask umount returned busy %s", ap->path);
+		return 1;
+	}
+
+	ioctl(ap->ioctlfd, AUTOFS_IOC_CATATONIC, 0);
+	close(ap->ioctlfd);
+	ap->ioctlfd = -1;
 	close(ap->state_pipe[0]);
 	close(ap->state_pipe[1]);
 	ap->state_pipe[0] = -1;
@@ -263,45 +272,11 @@ void close_mount_fds(struct autofs_point *ap)
 	if (ap->kpipefd >= 0)
 		close(ap->kpipefd);
 
-	return;
-}
-
-int umount_autofs_indirect(struct autofs_point *ap, const char *root)
-{
-	struct ioctl_ops *ops = get_ioctl_ops();
-	char buf[MAX_ERR_BUF];
-	char mountpoint[PATH_MAX + 1];
-	int rv, retries;
-	unsigned int unused;
-
-	if (root)
-		strcpy(mountpoint, root);
-	else
-		strcpy(mountpoint, ap->path);
-
-	/* If we are trying to shutdown make sure we can umount */
-	rv = ops->askumount(ap->logopt, ap->ioctlfd, &unused);
-	if (rv == -1) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr("ioctl failed: %s", estr);
-		return 1;
-	} else if (!unused) {
-#if defined(ENABLE_IGNORE_BUSY_MOUNTS) || defined(ENABLE_FORCED_SHUTDOWN)
-		if (!ap->shutdown)
-			return 1;
-		error(ap->logopt, "ask umount returned busy %s", ap->path);
-#else
-		return 1;
-#endif
-	}
-
-	ops->close(ap->logopt, ap->ioctlfd);
-	ap->ioctlfd = -1;
 	sched_yield();
 
 	retries = UMOUNT_RETRIES;
-	while ((rv = umount(mountpoint)) == -1 && retries--) {
-		struct timespec tm = {0, 200000000};
+	while ((rv = umount(ap->path)) == -1 && retries--) {
+		struct timespec tm = {0, 100000000};
 		if (errno != EBUSY)
 			break;
 		nanosleep(&tm, NULL);
@@ -312,71 +287,79 @@ int umount_autofs_indirect(struct autofs_point *ap, const char *root)
 		case ENOENT:
 		case EINVAL:
 			error(ap->logopt,
-			      "mount point %s does not exist", mountpoint);
-			close_mount_fds(ap);
+			      "mount point %s does not exist", ap->path);
 			return 0;
 			break;
 		case EBUSY:
-			debug(ap->logopt,
-			      "mount point %s is in use", mountpoint);
-			if (ap->state == ST_SHUTDOWN_FORCE) {
-				close_mount_fds(ap);
+			error(ap->logopt,
+			      "mount point %s is in use", ap->path);
+			if (ap->state == ST_SHUTDOWN_FORCE)
 				goto force_umount;
-			} else {
-				/*
-				 * If the umount returns EBUSY there may be
-				 * a mount request in progress so we need to
-				 * recover unless we have been explicitly
-				 * asked to shutdown and configure option
-				 * ENABLE_IGNORE_BUSY_MOUNTS is enabled.
-				 */
-#ifdef ENABLE_IGNORE_BUSY_MOUNTS
-				if (ap->shutdown) {
-					close_mount_fds(ap);
-					return 0;
-				}
-#endif
-				ops->open(ap->logopt,
-					  &ap->ioctlfd, ap->dev, mountpoint);
-				if (ap->ioctlfd < 0) {
-					warn(ap->logopt,
-					     "could not recover autofs path %s",
-					     mountpoint);
-					close_mount_fds(ap);
-					return 0;
-				}
-			}
+			else
+				return 0;
 			break;
 		case ENOTDIR:
 			error(ap->logopt, "mount point is not a directory");
-			close_mount_fds(ap);
 			return 0;
 			break;
 		}
 		return 1;
 	}
 
-	/*
-	 * We have successfully umounted the mount so we now close
-	 * the descriptors. The kernel end of the kernel pipe will
-	 * have been put during the umount super block cleanup.
-	 */
-	close_mount_fds(ap);
-
 force_umount:
 	if (rv != 0) {
 		warn(ap->logopt,
-		     "forcing umount of indirect mount %s", mountpoint);
-		rv = umount2(mountpoint, MNT_DETACH);
+		     "forcing umount of indirect mount %s", ap->path);
+		rv = umount2(ap->path, MNT_DETACH);
 	} else {
-		info(ap->logopt, "umounted indirect mount %s", mountpoint);
+		info(ap->logopt, "umounted indirect mount %s", ap->path);
 		if (ap->submount)
-			rm_unwanted(ap, mountpoint, 1);
+			rm_unwanted(ap->logopt, ap->path, 1, ap->dev);
 	}
 
 	return rv;
 }
 
+static int expire_indirect(struct autofs_point *ap, int ioctlfd, const char *path, unsigned int when)
+{
+	char buf[MAX_ERR_BUF];
+	int ret, retries;
+	struct stat st;
+
+	if (fstat(ioctlfd, &st) == -1) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		error(ap->logopt, "fstat failed: %s", estr);
+		return 0;
+	}
+
+	retries = (count_mounts(ap->logopt, path, st.st_dev) + 1) * EXPIRE_RETRIES;
+
+	while (retries--) {
+		struct timespec tm = {0, 100000000};
+
+		/* Ggenerate expire message for the mount. */
+		ret = ioctl(ioctlfd, AUTOFS_IOC_EXPIRE_DIRECT, &when);
+		if (ret == -1) {
+			/* Mount has gone away */
+			if (errno == EBADF || errno == EINVAL)
+				return 1;
+
+			/* Other than need to wait for the kernel ? */
+			if (errno != EAGAIN)
+				return 0;
+		}
+
+		nanosleep(&tm, NULL);
+	}
+
+	if (!ioctl(ioctlfd, AUTOFS_IOC_ASKUMOUNT, &ret)) {
+		if (!ret)
+			return 0;
+	}
+
+	return 1;
+}
+
 static void mnts_cleanup(void *arg)
 {
 	struct mnt_list *mnts = (struct mnt_list *) arg;
@@ -386,7 +369,6 @@ static void mnts_cleanup(void *arg)
 
 void *expire_proc_indirect(void *arg)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
 	struct autofs_point *ap;
 	struct mapent *me = NULL;
 	struct mnt_list *mnts = NULL, *next;
@@ -394,7 +376,6 @@ void *expire_proc_indirect(void *arg)
 	struct expire_args ec;
 	unsigned int now;
 	int offsets, submnts, count;
-	int retries;
 	int ioctlfd, cur_state;
 	int status, ret, left;
 
@@ -436,62 +417,8 @@ void *expire_proc_indirect(void *arg)
 			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 			if (strstr(next->opts, "indirect"))
 				master_notify_submount(ap, next->path, ap->state);
-			else if (strstr(next->opts, "offset")) {
-				struct map_source *map;
-				struct mapent_cache *mc = NULL;
-				struct mapent *me = NULL;
-				struct stat st;
-
-				/* It's got a mount, deal with in the outer loop */
-				if (is_mounted(_PATH_MOUNTED, next->path, MNTS_REAL)) {
-					pthread_setcancelstate(cur_state, NULL);
-					continue;
-				}
-
-				/* Don't touch submounts */
-				if (master_find_submount(ap, next->path)) {
-					pthread_setcancelstate(cur_state, NULL);
-					continue;
-				}
-
-				master_source_writelock(ap->entry);
-
-				map = ap->entry->maps;
-				while (map) {
-					mc = map->mc;
-					cache_writelock(mc);
-					me = cache_lookup_distinct(mc, next->path);
-					if (me)
-						break;
-					cache_unlock(mc);
-					map = map->next;
-				}
-
-				if (!mc || !me) {
-					master_source_unlock(ap->entry);
-					pthread_setcancelstate(cur_state, NULL);
-					continue;
-				}
-
-				if (me->ioctlfd == -1) {
-					cache_unlock(mc);
-					master_source_unlock(ap->entry);
-					pthread_setcancelstate(cur_state, NULL);
-					continue;
-				}
-
-				/* Check for manual umount */
-				if (fstat(me->ioctlfd, &st) == -1 ||
-				    !count_mounts(ap, me->key, st.st_dev)) {
-					ops->close(ap->logopt, me->ioctlfd);
-					me->ioctlfd = -1;
-				}
-
-				cache_unlock(mc);
-				master_source_unlock(ap->entry);
-			}
-
 			pthread_setcancelstate(cur_state, NULL);
+
 			continue;
 		}
 
@@ -513,43 +440,41 @@ void *expire_proc_indirect(void *arg)
 		 * Otherwise it's a top level indirect mount (possibly
 		 * with offsets in it) and we use the usual ioctlfd.
 		 */
-		pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
-		master_source_readlock(ap->entry);
 		me = lookup_source_mapent(ap, next->path, LKP_DISTINCT);
 		if (!me && ind_key)
 			me = lookup_source_mapent(ap, ind_key, LKP_NORMAL);
-		pthread_cleanup_pop(1);
+		if (!me)
+			continue;
 
-		ioctlfd = ap->ioctlfd;
-		if (me) {
-			if (*me->key == '/')
-				ioctlfd = me->ioctlfd;
-			cache_unlock(me->mc);
+		if (*me->key == '/') {
+			ioctlfd = me->ioctlfd;
+		} else {
+			ioctlfd = ap->ioctlfd;
 		}
+		cache_unlock(me->mc);
 
 		debug(ap->logopt, "expire %s", next->path);
 
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		ret = ops->expire(ap->logopt, ioctlfd, next->path, now);
-		if (ret)
+		ret = expire_indirect(ap, ioctlfd, next->path, now);
+		if (!ret)
 			left++;
 		pthread_setcancelstate(cur_state, NULL);
 	}
+	pthread_cleanup_pop(1);
 
 	/*
 	 * If there are no more real mounts left we could still
-	 * have some offset mounts with no '/' offset or symlinks
-	 * so we need to umount or unlink them here.
+	 * have some offset mounts with no '/' offset so we need to
+	 * umount them here.
 	 */
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-	retries = (count_mounts(ap, ap->path, ap->dev) + 1);
-	while (retries--) {
-		ret = ops->expire(ap->logopt, ap->ioctlfd, ap->path, now);
-		if (ret)
+	if (mnts) {
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+		ret = expire_indirect(ap, ap->ioctlfd, ap->path, now);
+		if (!ret)
 			left++;
+		pthread_setcancelstate(cur_state, NULL);
 	}
-	pthread_setcancelstate(cur_state, NULL);
-	pthread_cleanup_pop(1);
 
 	count = offsets = submnts = 0;
 	mnts = get_mnt_list(_PROC_MOUNTS, ap->path, 0);
@@ -568,8 +493,8 @@ void *expire_proc_indirect(void *arg)
 	pthread_cleanup_pop(1);
 
 	if (submnts)
-		info(ap->logopt,
-		     "%d submounts remaining in %s", submnts, ap->path);
+		debug(ap->logopt,
+		      "%d submounts remaining in %s", submnts, ap->path);
 
 	/* 
 	 * EXPIRE_MULTI is synchronous, so we can be sure (famous last
@@ -587,61 +512,81 @@ void *expire_proc_indirect(void *arg)
 	return NULL;
 }
 
+static void pending_cond_destroy(void *arg)
+{
+	struct pending_args *mt;
+	int status;
+
+	mt = (struct pending_args *) arg;
+	status = pthread_cond_destroy(&mt->cond);
+	if (status)
+		fatal(status);
+}
+
 static void expire_send_fail(void *arg)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *mt = arg;
-	struct autofs_point *ap = mt->ap;
-	ops->send_fail(ap->logopt,
-		       ap->ioctlfd, mt->wait_queue_token, -ENOENT);
+	send_fail(mt->ap->logopt, mt->ap->ioctlfd, mt->wait_queue_token);
+}
+
+static void free_pending_args(void *arg)
+{
+	struct pending_args *mt = arg;
+	free(mt);
+}
+
+static void expire_mutex_unlock(void *arg)
+{
+	int status = pthread_mutex_unlock(&ea_mutex);
+	if (status)
+		fatal(status);
 }
 
 static void *do_expire_indirect(void *arg)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
-	struct pending_args *args, mt;
+	struct pending_args *mt;
 	struct autofs_point *ap;
 	int status, state;
 
-	args = (struct pending_args *) arg;
+	mt = (struct pending_args *) arg;
 
-	pending_mutex_lock(args);
-
-	memcpy(&mt, args, sizeof(struct pending_args));
-
-	ap = mt.ap;
-
-	args->signaled = 1;
-	status = pthread_cond_signal(&args->cond);
+	status = pthread_mutex_lock(&ea_mutex);
 	if (status)
 		fatal(status);
 
-	pending_mutex_unlock(args);
+	ap = mt->ap;
 
-	pthread_cleanup_push(expire_send_fail, &mt);
+	mt->signaled = 1;
+	status = pthread_cond_signal(&mt->cond);
+	if (status)
+		fatal(status);
 
-	status = do_expire(mt.ap, mt.name, mt.len);
+	expire_mutex_unlock(NULL);
+
+	pthread_cleanup_push(free_pending_args, mt);
+	pthread_cleanup_push(pending_cond_destroy, mt);
+	pthread_cleanup_push(expire_send_fail, mt);
+
+	status = do_expire(mt->ap, mt->name, mt->len);
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 	if (status)
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, mt.wait_queue_token, -status);
+		send_fail(ap->logopt, ap->ioctlfd, mt->wait_queue_token);
 	else
-		ops->send_ready(ap->logopt,
-				ap->ioctlfd, mt.wait_queue_token);
+		send_ready(ap->logopt, ap->ioctlfd, mt->wait_queue_token);
 	pthread_setcancelstate(state, NULL);
 
 	pthread_cleanup_pop(0);
+	pthread_cleanup_pop(1);
+	pthread_cleanup_pop(1);
 
 	return NULL;
 }
 
 int handle_packet_expire_indirect(struct autofs_point *ap, autofs_packet_expire_indirect_t *pkt)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *mt;
 	char buf[MAX_ERR_BUF];
 	pthread_t thid;
-	struct timespec wait;
 	int status, state;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
@@ -653,15 +598,16 @@ int handle_packet_expire_indirect(struct autofs_point *ap, autofs_packet_expire_
 	if (!mt) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		logerr("malloc: %s", estr);
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, pkt->wait_queue_token, -ENOMEM);
+		send_fail(ap->logopt, ap->ioctlfd, pkt->wait_queue_token);
 		pthread_setcancelstate(state, NULL);
 		return 1;
 	}
 
-	pending_cond_init(mt);
+	status = pthread_cond_init(&mt->cond, NULL);
+	if (status)
+		fatal(status);
 
-	status = pthread_mutex_init(&mt->mutex, NULL);
+	status = pthread_mutex_lock(&ea_mutex);
 	if (status)
 		fatal(status);
 
@@ -671,39 +617,27 @@ int handle_packet_expire_indirect(struct autofs_point *ap, autofs_packet_expire_
 	mt->len = pkt->len;
 	mt->wait_queue_token = pkt->wait_queue_token;
 
-	pending_mutex_lock(mt);
-
-	status = pthread_create(&thid, &th_attr_detached, do_expire_indirect, mt);
+	status = pthread_create(&thid, &thread_attr, do_expire_indirect, mt);
 	if (status) {
 		error(ap->logopt, "expire thread create failed");
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, pkt->wait_queue_token, -status);
-		pending_mutex_unlock(mt);
+		send_fail(ap->logopt, ap->ioctlfd, pkt->wait_queue_token);
+		expire_mutex_unlock(NULL);
 		pending_cond_destroy(mt);
-		pending_mutex_destroy(mt);
 		free_pending_args(mt);
 		pthread_setcancelstate(state, NULL);
 		return 1;
 	}
 
-	pthread_cleanup_push(free_pending_args, mt);
-	pthread_cleanup_push(pending_mutex_destroy, mt);
-	pthread_cleanup_push(pending_cond_destroy, mt);
-	pthread_cleanup_push(pending_mutex_unlock, mt);
+	pthread_cleanup_push(expire_mutex_unlock, NULL);
 	pthread_setcancelstate(state, NULL);
 
 	mt->signaled = 0;
 	while (!mt->signaled) {
-		clock_gettime(CLOCK_MONOTONIC, &wait);
-		wait.tv_sec += 2;
-		status = pthread_cond_timedwait(&mt->cond, &mt->mutex, &wait);
-		if (status && status != ETIMEDOUT)
+		status = pthread_cond_wait(&mt->cond, &ea_mutex);
+		if (status)
 			fatal(status);
 	}
 
-	pthread_cleanup_pop(1);
-	pthread_cleanup_pop(1);
-	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
 
 	return 0;
@@ -711,58 +645,66 @@ int handle_packet_expire_indirect(struct autofs_point *ap, autofs_packet_expire_
 
 static void mount_send_fail(void *arg)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *mt = arg;
-	struct autofs_point *ap = mt->ap;
-	ops->send_fail(ap->logopt,
-		       ap->ioctlfd, mt->wait_queue_token, -ENOENT);
+	send_fail(mt->ap->logopt, mt->ap->ioctlfd, mt->wait_queue_token);
+}
+
+static void mount_mutex_unlock(void *arg)
+{
+	int status = pthread_mutex_unlock(&ma_mutex);
+	if (status)
+		fatal(status);
 }
 
 static void *do_mount_indirect(void *arg)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
-	struct pending_args *args, mt;
+	struct pending_args *mt;
 	struct autofs_point *ap;
 	char buf[PATH_MAX + 1];
 	struct stat st;
-	int len, status, state;
+	struct passwd pw;
+	struct passwd *ppw = &pw;
+	struct passwd **pppw = &ppw;
+	struct group gr;
+	struct group *pgr;
+	struct group **ppgr;
+	char *pw_tmp, *gr_tmp;
+	struct thread_stdenv_vars *tsv;
+	int len, tmplen, grplen, status, state;
 
-	args = (struct pending_args *) arg;
+	mt = (struct pending_args *) arg;
 
-	pending_mutex_lock(args);
-
-	memcpy(&mt, args, sizeof(struct pending_args));
-
-	ap = mt.ap;
-
-	set_thread_mount_request_log_id(&mt);
-
-	args->signaled = 1;
-	status = pthread_cond_signal(&args->cond);
+	status = pthread_mutex_lock(&ma_mutex);
 	if (status)
 		fatal(status);
 
-	pending_mutex_unlock(args);
+	ap = mt->ap;
+	mt->status = 0;
 
-	pthread_cleanup_push(mount_send_fail, &mt);
+	mt->signaled = 1;
+	status = pthread_cond_signal(&mt->cond);
+	if (status)
+		fatal(status);
+
+	mount_mutex_unlock(NULL);
+
+	pthread_cleanup_push(free_pending_args, mt);
+	pthread_cleanup_push(pending_cond_destroy, mt);
+	pthread_cleanup_push(mount_send_fail, mt);
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 
-	len = ncat_path(buf, sizeof(buf), ap->path, mt.name, mt.len);
+	len = ncat_path(buf, sizeof(buf), ap->path, mt->name, mt->len);
 	if (!len) {
 		crit(ap->logopt, "path to be mounted is to long");
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, mt.wait_queue_token,
-			      -ENAMETOOLONG);
 		pthread_setcancelstate(state, NULL);
 		pthread_exit(NULL);
 	}
 
 	status = lstat(buf, &st);
-	if (status != -1 && !(S_ISDIR(st.st_mode) && st.st_dev == mt.dev)) {
+	if (status != -1 && !(S_ISDIR(st.st_mode) && st.st_dev == mt->dev)) {
 		error(ap->logopt,
 		      "indirect trigger not valid or already mounted %s", buf);
-		ops->send_ready(ap->logopt, ap->ioctlfd, mt.wait_queue_token);
 		pthread_setcancelstate(state, NULL);
 		pthread_exit(NULL);
 	}
@@ -771,86 +713,181 @@ static void *do_mount_indirect(void *arg)
 
 	info(ap->logopt, "attempting to mount entry %s", buf);
 
-	set_tsd_user_vars(ap->logopt, mt.uid, mt.gid);
+	/*
+	 * Setup thread specific data values for macro
+	 * substution in map entries during the mount.
+	 * Best effort only as it must go ahead.
+	 */
 
-	status = lookup_nss_mount(ap, NULL, mt.name, mt.len);
+	tsv = malloc(sizeof(struct thread_stdenv_vars));
+	if (!tsv) 
+		goto cont;
+
+	tsv->uid = mt->uid;
+	tsv->gid = mt->gid;
+
+	/* Try to get passwd info */
+
+	tmplen = sysconf(_SC_GETPW_R_SIZE_MAX);
+	if (tmplen < 0) {
+		error(ap->logopt, "failed to get buffer size for getpwuid_r");
+		free(tsv);
+		goto cont;
+	}
+
+	pw_tmp = malloc(tmplen + 1);
+	if (!pw_tmp) {
+		error(ap->logopt, "failed to malloc buffer for getpwuid_r");
+		free(tsv);
+		goto cont;
+	}
+
+	status = getpwuid_r(mt->uid, ppw, pw_tmp, tmplen, pppw);
+	if (status || !ppw) {
+		error(ap->logopt, "failed to get passwd info from getpwuid_r");
+		free(tsv);
+		free(pw_tmp);
+		goto cont;
+	}
+
+	tsv->user = strdup(pw.pw_name);
+	if (!tsv->user) {
+		error(ap->logopt, "failed to malloc buffer for user");
+		free(tsv);
+		free(pw_tmp);
+		goto cont;
+	}
+
+	tsv->home = strdup(pw.pw_dir);
+	if (!tsv->user) {
+		error(ap->logopt, "failed to malloc buffer for home");
+		free(pw_tmp);
+		free(tsv->user);
+		free(tsv);
+		goto cont;
+	}
+
+	free(pw_tmp);
+
+	/* Try to get group info */
+
+	grplen = sysconf(_SC_GETGR_R_SIZE_MAX);
+	if (tmplen < 0) {
+		error(ap->logopt, "failed to get buffer size for getgrgid_r");
+		free(tsv->user);
+		free(tsv->home);
+		free(tsv);
+		goto cont;
+	}
+
+	gr_tmp = NULL;
+	tmplen = grplen;
+	while (1) {
+		char *tmp = realloc(gr_tmp, tmplen + 1);
+		if (!tmp) {
+			error(ap->logopt, "failed to malloc buffer for getgrgid_r");
+			if (gr_tmp)
+				free(gr_tmp);
+			free(tsv->user);
+			free(tsv->home);
+			free(tsv);
+			goto cont;
+		}
+		gr_tmp = tmp;
+		pgr = &gr;
+		ppgr = &pgr;
+		status = getgrgid_r(mt->gid, pgr, gr_tmp, tmplen, ppgr);
+		if (status != ERANGE)
+			break;
+		tmplen += grplen;
+	}
+
+	if (status || !pgr) {
+		error(ap->logopt, "failed to get group info from getgrgid_r");
+		free(tsv->user);
+		free(tsv->home);
+		free(tsv);
+		free(gr_tmp);
+		goto cont;
+	}
+
+	tsv->group = strdup(gr.gr_name);
+	if (!tsv->group) {
+		error(ap->logopt, "failed to malloc buffer for group");
+		free(tsv->user);
+		free(tsv->home);
+		free(tsv);
+		free(gr_tmp);
+		goto cont;
+	}
+
+	free(gr_tmp);
+
+	status = pthread_setspecific(key_thread_stdenv_vars, tsv);
+	if (status) {
+		error(ap->logopt, "failed to set stdenv thread var");
+		free(tsv->group);
+		free(tsv->user);
+		free(tsv->home);
+		free(tsv);
+	}
+cont:
+	status = lookup_nss_mount(ap, NULL, mt->name, mt->len);
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 	if (status) {
-		ops->send_ready(ap->logopt,
-				ap->ioctlfd, mt.wait_queue_token);
+		send_ready(ap->logopt, ap->ioctlfd, mt->wait_queue_token);
 		info(ap->logopt, "mounted %s", buf);
 	} else {
-		/* TODO: get mount return status from lookup_nss_mount */
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, mt.wait_queue_token, -ENOENT);
+		send_fail(ap->logopt, ap->ioctlfd, mt->wait_queue_token);
 		info(ap->logopt, "failed to mount %s", buf);
 	}
 	pthread_setcancelstate(state, NULL);
 
 	pthread_cleanup_pop(0);
+	pthread_cleanup_pop(1);
+	pthread_cleanup_pop(1);
 
 	return NULL;
 }
 
 int handle_packet_missing_indirect(struct autofs_point *ap, autofs_packet_missing_indirect_t *pkt)
 {
-	struct ioctl_ops *ops = get_ioctl_ops();
 	pthread_t thid;
 	char buf[MAX_ERR_BUF];
 	struct pending_args *mt;
-	struct timespec wait;
-	struct mapent *me;
 	int status, state;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 
-	master_mutex_lock();
-
 	debug(ap->logopt, "token %ld, name %s, request pid %u",
 		(unsigned long) pkt->wait_queue_token, pkt->name, pkt->pid);
 
 	/* Ignore packet if we're trying to shut down */
-	if (ap->shutdown || ap->state == ST_SHUTDOWN_FORCE) {
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, pkt->wait_queue_token, -ENOENT);
-		master_mutex_unlock();
+	if (ap->shutdown ||
+	    ap->state == ST_SHUTDOWN_FORCE ||
+	    ap->state == ST_SHUTDOWN) {
+		send_fail(ap->logopt, ap->ioctlfd, pkt->wait_queue_token);
 		pthread_setcancelstate(state, NULL);
 		return 0;
 	}
 
-	/* Check if we recorded a mount fail for this key anywhere */
-	me = lookup_source_mapent(ap, pkt->name, LKP_DISTINCT);
-	if (me) {
-		if (me->status >= monotonic_time(NULL)) {
-			ops->send_fail(ap->logopt, ap->ioctlfd,
-				       pkt->wait_queue_token, -ENOENT);
-			cache_unlock(me->mc);
-			master_mutex_unlock();
-			pthread_setcancelstate(state, NULL);
-			return 0;
-		}
-		cache_unlock(me->mc);
-	}
-
 	mt = malloc(sizeof(struct pending_args));
 	if (!mt) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		logerr("malloc: %s", estr);
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, pkt->wait_queue_token, -ENOMEM);
-		master_mutex_unlock();
+		send_fail(ap->logopt, ap->ioctlfd, pkt->wait_queue_token);
 		pthread_setcancelstate(state, NULL);
 		return 1;
 	}
 	memset(mt, 0, sizeof(struct pending_args));
 
-	pending_cond_init(mt);
-
-	status = pthread_mutex_init(&mt->mutex, NULL);
+	status = pthread_cond_init(&mt->cond, NULL);
 	if (status)
 		fatal(status);
 
-	pending_mutex_lock(mt);
+	status = pthread_mutex_lock(&ma_mutex);
+	if (status)
+		fatal(status);
 
 	mt->ap = ap;
 	strncpy(mt->name, pkt->name, pkt->len);
@@ -859,43 +896,29 @@ int handle_packet_missing_indirect(struct autofs_point *ap, autofs_packet_missin
 	mt->dev = pkt->dev;
 	mt->uid = pkt->uid;
 	mt->gid = pkt->gid;
-	mt->pid = pkt->pid;
 	mt->wait_queue_token = pkt->wait_queue_token;
 
-	status = pthread_create(&thid, &th_attr_detached, do_mount_indirect, mt);
+	status = pthread_create(&thid, &thread_attr, do_mount_indirect, mt);
 	if (status) {
 		error(ap->logopt, "expire thread create failed");
-		ops->send_fail(ap->logopt,
-			       ap->ioctlfd, pkt->wait_queue_token, -status);
-		master_mutex_unlock();
-		pending_mutex_unlock(mt);
+		send_fail(ap->logopt, ap->ioctlfd, pkt->wait_queue_token);
+		mount_mutex_unlock(NULL);
 		pending_cond_destroy(mt);
-		pending_mutex_destroy(mt);
 		free_pending_args(mt);
 		pthread_setcancelstate(state, NULL);
 		return 1;
 	}
 
-	master_mutex_unlock();
-
-	pthread_cleanup_push(free_pending_args, mt);
-	pthread_cleanup_push(pending_mutex_destroy, mt);
-	pthread_cleanup_push(pending_cond_destroy, mt);
-	pthread_cleanup_push(pending_mutex_unlock, mt);
+	pthread_cleanup_push(mount_mutex_unlock, NULL);
 	pthread_setcancelstate(state, NULL);
 
 	mt->signaled = 0;
 	while (!mt->signaled) {
-		clock_gettime(CLOCK_MONOTONIC, &wait);
-		wait.tv_sec += 2;
-		status = pthread_cond_timedwait(&mt->cond, &mt->mutex, &wait);
-		if (status && status != ETIMEDOUT)
+		status = pthread_cond_wait(&mt->cond, &ma_mutex);
+		if (status)
 			fatal(status);
 	}
 
-	pthread_cleanup_pop(1);
-	pthread_cleanup_pop(1);
-	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
 
 	return 0;
