@@ -13,6 +13,7 @@
  *
  * ----------------------------------------------------------------------- */
 
+#include <fcntl.h>
 #include <signal.h>
 #include <stdarg.h>
 #include <stdio.h>
@@ -20,8 +21,8 @@
 #include <string.h>
 #include <sys/types.h>
 #include <dirent.h>
+#include <unistd.h>
 #include <time.h>
-#include <poll.h>
 #include <sys/wait.h>
 #include <sys/stat.h>
 #include <sys/mount.h>
@@ -32,11 +33,11 @@ static pthread_mutex_t spawn_mutex = PTHREAD_MUTEX_INITIALIZER;
 
 #define SPAWN_OPT_NONE		0x0000
 #define SPAWN_OPT_LOCK		0x0001
-#define SPAWN_OPT_OPEN		0x0002
+#define SPAWN_OPT_ACCESS	0x0002
 
 #define MTAB_LOCK_RETRIES	3
 
-void dump_core(void)
+inline void dump_core(void)
 {
 	sigset_t segv;
 
@@ -88,37 +89,7 @@ void reset_signals(void)
 
 #define ERRBUFSIZ 2047		/* Max length of error string excl \0 */
 
-static int timed_read(int pipe, char *buf, size_t len, int time)
-{
-	struct pollfd pfd[1];
-	int timeout = time;
-	int ret;
-
-	pfd[0].fd = pipe;
-	pfd[0].events = POLLIN;
-
-	if (time != -1) {
-		if (time >= (INT_MAX - 1)/1000)
-			timeout = INT_MAX - 1;
-		else
-			timeout = time * 1000;
-	}
-
-	ret = poll(pfd, 1, timeout);
-	if (ret <= 0) {
-		if (ret == 0)
-			ret = -ETIMEDOUT;
-		return ret;
-	}
-
-	while ((ret = read(pipe, buf, len)) == -1 && errno == EINTR);
-
-	return ret;
-}
-
-static int do_spawn(unsigned logopt, unsigned int wait,
-		    unsigned int options, const char *prog,
-		    const char *const *argv)
+static int do_spawn(unsigned logopt, unsigned int options, const char *prog, const char *const *argv)
 {
 	pid_t f;
 	int ret, status, pipefd[2];
@@ -126,13 +97,13 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 	int errp, errn;
 	int cancel_state;
 	unsigned int use_lock = options & SPAWN_OPT_LOCK;
-	unsigned int use_open = options & SPAWN_OPT_OPEN;
+	unsigned int use_access = options & SPAWN_OPT_ACCESS;
 	sigset_t allsigs, tmpsig, oldsig;
 	struct thread_stdenv_vars *tsv;
 	pid_t euid = 0;
 	gid_t egid = 0;
 
-	if (open_pipe(pipefd))
+	if (pipe(pipefd))
 		return -1;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel_state);
@@ -154,35 +125,23 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 
 	f = fork();
 	if (f == 0) {
-		char **pargv = (char **) argv;
-		int loc = 0;
-
 		reset_signals();
 		close(pipefd[0]);
 		dup2(pipefd[1], STDOUT_FILENO);
 		dup2(pipefd[1], STDERR_FILENO);
 		close(pipefd[1]);
 
-		/* what to mount must always be second last */
-		while (*pargv++)
-			loc++;
-		if (loc <= 3)
-			goto done;
-		loc -= 2;
-
-		/*
-		 * If the mount location starts with a "/" then it is
-		 * a local path. In this case it is a bind mount, a
-		 * loopback mount or a file system that uses a local
-		 * path so we need to check for dependent mounts.
-		 *
-		 * I hope host names are never allowed "/" as first char
-		 */
-		if (use_open && *(argv[loc]) == '/') {
-			int fd;
-
+		/* Bind mount - check target exists */
+		if (use_access) {
+			char **pargv = (char **) argv;
+			int argc = 0;
 			pid_t pgrp = getpgrp();
 
+			/* what to mount must always be second last */
+			while (*pargv++)
+				argc++;
+			argc -= 2;
+
 			/*
 			 * Pretend to be requesting user and set non-autofs
 			 * program group to trigger mount
@@ -193,24 +152,15 @@ static int do_spawn(unsigned logopt, unsigned int wait,
 			}
 			setpgrp();
 
-			/*
-			 * Trigger the recursive mount.
-			 *
-			 * Ignore the open(2) return code as there may be
-			 * multiple waiters for this mount and we need to
-			 * let the VFS handle returns to each individual
-			 * waiter.
-			 */
-			fd = open(argv[loc], O_DIRECTORY);
-			if (fd != -1)
-				close(fd);
+			/* Trigger the recursive mount */
+			if (access(argv[argc], F_OK) == -1)
+				_exit(errno);
 
 			seteuid(0);
 			setegid(0);
-			if (pgrp >= 0)
-				setpgid(0, pgrp);
+			setpgid(0, pgrp);
 		}
-done:
+
 		execv(prog, (char *const *) argv);
 		_exit(255);	/* execv() failed */
 	} else {
@@ -235,8 +185,10 @@ done:
 
 		errp = 0;
 		do {
-			errn = timed_read(pipefd[0],
-					  errbuf + errp, ERRBUFSIZ - errp, wait);
+			while ((errn =
+				read(pipefd[0], errbuf + errp, ERRBUFSIZ - errp)) == -1
+			       && errno == EINTR);
+
 			if (errn > 0) {
 				errp += errn;
 
@@ -261,9 +213,6 @@ done:
 			}
 		} while (errn > 0);
 
-		if (errn == -ETIMEDOUT)
-			kill(f, SIGTERM);
-
 		close(pipefd[0]);
 
 		if (errp > 0) {
@@ -289,7 +238,7 @@ done:
 
 int spawnv(unsigned logopt, const char *prog, const char *const *argv)
 {
-	return do_spawn(logopt, -1, SPAWN_OPT_NONE, prog, argv);
+	return do_spawn(logopt, SPAWN_OPT_NONE, prog, argv);
 }
 
 int spawnl(unsigned logopt, const char *prog, ...)
@@ -310,7 +259,7 @@ int spawnl(unsigned logopt, const char *prog, ...)
 	while ((*p++ = va_arg(arg, char *)));
 	va_end(arg);
 
-	return do_spawn(logopt, -1, SPAWN_OPT_NONE, prog, (const char **) argv);
+	return do_spawn(logopt, SPAWN_OPT_NONE, prog, (const char **) argv);
 }
 
 int spawn_mount(unsigned logopt, ...)
@@ -320,70 +269,45 @@ int spawn_mount(unsigned logopt, ...)
 	char **argv, **p;
 	char prog[] = PATH_MOUNT;
 	char arg0[] = PATH_MOUNT;
-	char argn[] = "-n";
 	/* In case we need to use the fake option to mount */
 	char arg_fake[] = "-f";
 	unsigned int options;
 	unsigned int retries = MTAB_LOCK_RETRIES;
-	int update_mtab = 1, ret, printed = 0;
-	unsigned int wait = defaults_get_mount_wait();
-	char buf[PATH_MAX + 1];
+	int ret, printed = 0;
 
 	/* If we use mount locking we can't validate the location */
 #ifdef ENABLE_MOUNT_LOCKING
 	options = SPAWN_OPT_LOCK;
 #else
-	options = SPAWN_OPT_OPEN;
+	options = SPAWN_OPT_NONE;
 #endif
 
 	va_start(arg, logopt);
 	for (argc = 1; va_arg(arg, char *); argc++);
 	va_end(arg);
 
-	ret = readlink(_PATH_MOUNTED, buf, PATH_MAX);
-	if (ret != -1) {
-		buf[ret] = '\0';
-		if (!strcmp(buf, _PROC_MOUNTS) ||
-		    !strcmp(buf, _PROC_SELF_MOUNTS)) {
-			debug(logopt,
-			      "mtab link detected, passing -n to mount");
-			argc++;
-			update_mtab = 0;
-		}
-	}
-
 	/* Alloc 1 extra slot in case we need to use the "-f" option */
-	if (!(argv = alloca(sizeof(char *) * (argc + 2))))
+	if (!(argv = alloca(sizeof(char *) * argc + 2)))
 		return -1;
 
 	argv[0] = arg0;
 
 	va_start(arg, logopt);
-	if (update_mtab)
-		p = argv + 1;
-	else {
-		argv[1] = argn;
-		p = argv + 2;
-	}
+	p = argv + 1;
 	while ((*p = va_arg(arg, char *))) {
-		if (options == SPAWN_OPT_OPEN && !strcmp(*p, "-t")) {
+		if (options == SPAWN_OPT_NONE && !strcmp(*p, "-o")) {
 			*(++p) = va_arg(arg, char *);
 			if (!*p)
 				break;
-			/*
-			 * A cifs mount location begins with a "/" but
-			 * is not a local path, so don't try to resolve
-			 * it. Mmmm ... does anyone use smbfs these days?
-			 */
-			if (strstr(*p, "cifs"))
-				options = SPAWN_OPT_NONE;
+			if (strstr(*p, "loop"))
+				options = SPAWN_OPT_ACCESS;
 		}
 		p++;
 	}
 	va_end(arg);
 
 	while (retries--) {
-		ret = do_spawn(logopt, wait, options, prog, (const char **) argv);
+		ret = do_spawn(logopt, options, prog, (const char **) argv);
 		if (ret & MTAB_NOTUPDATED) {
 			struct timespec tm = {3, 0};
 
@@ -435,7 +359,7 @@ int spawn_mount(unsigned logopt, ...)
 
 /*
  * For bind mounts that depend on the target being mounted (possibly
- * itself an automount) we attempt to mount the target using an open(2)
+ * itself an automount) we attempt to mount the target using an access
  * call. For this to work the location must be the second last arg.
  *
  * NOTE: If mount locking is enabled this type of recursive mount cannot
@@ -449,19 +373,17 @@ int spawn_bind_mount(unsigned logopt, ...)
 	char prog[] = PATH_MOUNT;
 	char arg0[] = PATH_MOUNT;
 	char bind[] = "--bind";
-	char argn[] = "-n";
 	/* In case we need to use the fake option to mount */
 	char arg_fake[] = "-f";
 	unsigned int options;
 	unsigned int retries = MTAB_LOCK_RETRIES;
-	int update_mtab = 1, ret, printed = 0;
-	char buf[PATH_MAX + 1];
+	int ret, printed = 0;
 
 	/* If we use mount locking we can't validate the location */
 #ifdef ENABLE_MOUNT_LOCKING
 	options = SPAWN_OPT_LOCK;
 #else
-	options = SPAWN_OPT_OPEN;
+	options = SPAWN_OPT_ACCESS;
 #endif
 
 	/*
@@ -472,36 +394,19 @@ int spawn_bind_mount(unsigned logopt, ...)
 	for (argc = 2; va_arg(arg, char *); argc++);
 	va_end(arg);
 
-	ret = readlink(_PATH_MOUNTED, buf, PATH_MAX);
-	if (ret != -1) {
-		buf[ret] = '\0';
-		if (!strcmp(buf, _PROC_MOUNTS) ||
-		    !strcmp(buf, _PROC_SELF_MOUNTS)) {
-			debug(logopt,
-			      "mtab link detected, passing -n to mount");
-			argc++;
-			update_mtab = 0;
-		}
-	}
-
-	if (!(argv = alloca(sizeof(char *) * (argc + 2))))
+	if (!(argv = alloca(sizeof(char *) * argc + 2)))
 		return -1;
 
 	argv[0] = arg0;
 	argv[1] = bind;
 
 	va_start(arg, logopt);
-	if (update_mtab)
-		p = argv + 2;
-	else {
-		argv[2] = argn;
-		p = argv + 3;
-	}
+	p = argv + 2;
 	while ((*p++ = va_arg(arg, char *)));
 	va_end(arg);
 
 	while (retries--) {
-		ret = do_spawn(logopt, -1, options, prog, (const char **) argv);
+		ret = do_spawn(logopt, options, prog, (const char **) argv);
 		if (ret & MTAB_NOTUPDATED) {
 			struct timespec tm = {3, 0};
 
@@ -558,12 +463,9 @@ int spawn_umount(unsigned logopt, ...)
 	char **argv, **p;
 	char prog[] = PATH_UMOUNT;
 	char arg0[] = PATH_UMOUNT;
-	char argn[] = "-n";
 	unsigned int options;
 	unsigned int retries = MTAB_LOCK_RETRIES;
-	int update_mtab = 1, ret, printed = 0;
-	unsigned int wait = defaults_get_umount_wait();
-	char buf[PATH_MAX + 1];
+	int ret, printed = 0;
 
 #ifdef ENABLE_MOUNT_LOCKING
 	options = SPAWN_OPT_LOCK;
@@ -575,35 +477,18 @@ int spawn_umount(unsigned logopt, ...)
 	for (argc = 1; va_arg(arg, char *); argc++);
 	va_end(arg);
 
-	ret = readlink(_PATH_MOUNTED, buf, PATH_MAX);
-	if (ret != -1) {
-		buf[ret] = '\0';
-		if (!strcmp(buf, _PROC_MOUNTS) ||
-		    !strcmp(buf, _PROC_SELF_MOUNTS)) {
-			debug(logopt,
-			      "mtab link detected, passing -n to mount");
-			argc++;
-			update_mtab = 0;
-		}
-	}
-
 	if (!(argv = alloca(sizeof(char *) * argc + 1)))
 		return -1;
 
 	argv[0] = arg0;
 
 	va_start(arg, logopt);
-	if (update_mtab)
-		p = argv + 1;
-	else {
-		argv[1] = argn;
-		p = argv + 2;
-	}
+	p = argv + 1;
 	while ((*p++ = va_arg(arg, char *)));
 	va_end(arg);
 
 	while (retries--) {
-		ret = do_spawn(logopt, wait, options, prog, (const char **) argv);
+		ret = do_spawn(logopt, options, prog, (const char **) argv);
 		if (ret & MTAB_NOTUPDATED) {
 			/*
 			 * If the mount succeeded but the mtab was not
