@@ -16,8 +16,11 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <malloc.h>
+#include <fcntl.h>
+#include <unistd.h>
 #include <string.h>
 #include <signal.h>
+#include <alloca.h>
 #include <sys/param.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -29,7 +32,7 @@
 #define MODPREFIX "mount(autofs): "
 
 /* Attribute to create detached thread */
-extern pthread_attr_t th_attr_detached;
+extern pthread_attr_t thread_attr;
 extern struct startup_cond suc;
 
 int mount_version = AUTOFS_MOUNT_VERSION;	/* Required by protocol */
@@ -43,50 +46,45 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 		int name_len, const char *what, const char *fstype,
 		const char *c_options, void *context)
 {
-	struct startup_cond suc;
 	pthread_t thid;
-	char realpath[PATH_MAX];
-	char mountpoint[PATH_MAX];
+	char *fullpath;
 	const char **argv;
-	int argc, status;
-	int nobind = ap->flags & MOUNT_FLAG_NOBIND;
-	int ghost = ap->flags & MOUNT_FLAG_GHOST;
-	int symlnk = ap->flags & MOUNT_FLAG_SYMLINK;
-	time_t timeout = ap->entry->maps->exp_timeout;
+	int argc, status, ghost = ap->ghost;
+	time_t timeout = ap->exp_timeout;
 	unsigned logopt = ap->logopt;
-	struct map_type_info *info;
+	char *type, *format, *tmp, *tmp2;
 	struct master *master;
 	struct master_mapent *entry;
 	struct map_source *source;
 	struct autofs_point *nap;
 	char buf[MAX_ERR_BUF];
 	char *options, *p;
-	int len, ret;
-	int hosts = 0;
+	int ret;
+
+	fullpath = alloca(strlen(root) + name_len + 2);
+	if (!fullpath) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "alloca: %s", estr);
+		return 1;
+	}
 
 	/* Root offset of multi-mount */
-	len = strlen(root);
-	if (root[len - 1] == '/') {
-		strcpy(realpath, ap->path);
-		strcat(realpath, "/");
-		strcat(realpath, name);
-		len--;
-		strncpy(mountpoint, root, len);
-		mountpoint[len] = '\0';
-	} else if (*name == '/') {
-		if (ap->flags & MOUNT_FLAG_REMOUNT) {
-			strcpy(mountpoint, name);
-			strcpy(realpath, name);
-		} else {
-			strcpy(mountpoint, root);
-			strcpy(realpath, name);
-		}
-	} else {
-		strcpy(mountpoint, root);
-		strcat(mountpoint, "/");
-		strcpy(realpath, mountpoint);
-		strcat(mountpoint, name);
-		strcat(realpath, name);
+	if (*name == '/' && name_len == 1)
+		strcpy(fullpath, root);
+	else if (*name == '/')
+		strcpy(fullpath, name);
+	else {
+		strcpy(fullpath, root);
+		strcat(fullpath, "/");
+		strcat(fullpath, name);
+	}
+
+	if (is_mounted(_PATH_MOUNTED, fullpath, MNTS_REAL)) {
+		error(ap->logopt,
+		      MODPREFIX 
+		      "warning: about to mount over %s, continuing",
+		      fullpath);
+		return 0;
 	}
 
 	options = NULL;
@@ -116,20 +114,14 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 			while (*comma != '\0' && *comma != ',')
 				comma++;
 
-			if (_strncmp(cp, "nobrowse", 8) == 0)
+			if (strncmp(cp, "nobrowse", 8) == 0)
 				ghost = 0;
-			else if (_strncmp(cp, "nobind", 6) == 0)
-				nobind = 1;
-			else if (_strncmp(cp, "browse", 6) == 0)
+			else if (strncmp(cp, "browse", 6) == 0)
 				ghost = 1;
-			else if (_strncmp(cp, "symlink", 7) == 0)
-				symlnk = 1;
-			else if (_strncmp(cp, "hosts", 5) == 0)
-				hosts = 1;
-			else if (_strncmp(cp, "timeout=", 8) == 0) {
+			else if (strncmp(cp, "timeout=", 8) == 0) {
 				char *val = strchr(cp, '=');
 				unsigned tout;
-				if (val) {
+				if (val++) {
 					int ret = sscanf(cp, "timeout=%u", &tout);
 					if (ret)
 						timeout = tout;
@@ -143,19 +135,19 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 	}
 
 	debug(ap->logopt,
-	      MODPREFIX "mountpoint=%s what=%s options=%s",
-	      mountpoint, what, options);
+	      MODPREFIX "fullpath=%s what=%s options=%s",
+	      fullpath, what, options);
 
 	master = ap->entry->master;
 
-	entry = master_new_mapent(master, realpath, ap->entry->age);
+	entry = master_new_mapent(master, fullpath, ap->entry->age);
 	if (!entry) {
 		error(ap->logopt,
 		      MODPREFIX "failed to malloc master_mapent struct");
 		return 1;
 	}
 
-	ret = master_add_autofs_point(entry, logopt, nobind, ghost, 1);
+	ret = master_add_autofs_point(entry, timeout, logopt, ghost, 1);
 	if (!ret) {
 		error(ap->logopt,
 		      MODPREFIX "failed to add autofs_point to entry");
@@ -164,13 +156,8 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 	}
 	nap = entry->ap;
 	nap->parent = ap;
-	if (symlnk)
-		nap->flags |= MOUNT_FLAG_SYMLINK;
 
-	if (hosts)
-		argc = 0;
-	else
-		argc = 1;
+	argc = 1;
 
 	if (options) {
 		char *t = options;
@@ -182,51 +169,23 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 	}
 	argv = (const char **) alloca((argc + 1) * sizeof(char *));
 
-	if (hosts)
-		argc = 0;
-	else
-		argc = 1;
+	argc = 1;
 
-	/*
-	 * If a mount of a hosts map is being requested it will come
-	 * ro us via the options. Catch that below when processing the
-	 * option and create type info struct then.
-	 */
-	if (hosts)
-		info = parse_map_type_info("hosts:");
-	else
-		info = parse_map_type_info(what);
-	if (!info) {
-		error(ap->logopt, MODPREFIX "failed to parse map info");
-		master_free_mapent(entry);
-		return 1;
-	}
-	if (info->map)
-		argv[0] = info->map;
-	/*
-	 * If the parent map format is amd and the format isn't
-	 * specified in the map entry set it from the parent map
-	 * source.
-	 */
-	if (!info->format && ap->entry->maps) {
-		struct map_source *s = ap->entry->maps;
-		/*
-		 * For amd maps, if the format and source type aren't
-		 * specified try and set them from the parent.
-		 */
-		if (s->flags & MAP_FLAG_FORMAT_AMD) {
-			info->format = strdup("amd");
-			if (!info->format)
-				warn(ap->logopt, MODPREFIX
-				     "failed to set amd map format");
-			if (!info->type && s->type) {
-				info->type = strdup(s->type);
-				if (!info->type)
-					warn(ap->logopt, MODPREFIX
-					     "failed to set amd map type");
-			}
+	type = NULL;
+	format = NULL;
+
+	tmp = strchr(what, ':');
+	if (tmp) {
+		*tmp++ = '\0';
+		tmp2 = strchr(what, ',');
+		if (tmp2) {
+			*tmp2++ = '\0';
+			format = tmp2;
 		}
-	}
+		type = (char *) what;
+		argv[0] = tmp;
+	} else
+		argv[0] = (char *) what;
 
 	if (options) {
 		p = options;
@@ -240,85 +199,79 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name,
 	}
 	argv[argc] = NULL;
 
-	source = master_add_map_source(entry,
-				       info->type, info->format,
-				       time(NULL), argc, argv);
+	source = master_add_map_source(entry, type, format, time(NULL), argc, argv);
 	if (!source) {
 		error(ap->logopt,
 		      MODPREFIX "failed to add map source to entry");
 		master_free_mapent(entry);
-		free_map_type_info(info);
 		return 1;
 	}
-	free_map_type_info(info);
-	source->exp_timeout = timeout;
+
+	source->mc = cache_init(entry->ap, source);
+	if (!source->mc) {
+		error(ap->logopt, MODPREFIX "failed to init source cache");
+		master_free_mapent(entry);
+		return 1;
+	}
 
 	mounts_mutex_lock(ap);
 
-	if (source->flags & MAP_FLAG_FORMAT_AMD) {
-		struct amd_entry *am_entry = __master_find_amdmount(ap, entry->path);
-		if (am_entry && am_entry->pref) {
-			nap->pref = am_entry->pref;
-			am_entry->pref = NULL;
-		}
-		/* amd mounts don't support browse mode */
-		nap->flags &= ~MOUNT_FLAG_GHOST;
-	}
-
-	if (handle_mounts_startup_cond_init(&suc)) {
-		crit(ap->logopt, MODPREFIX
-		     "failed to init startup cond for mount %s", entry->path);
-		mounts_mutex_unlock(ap);
-		master_free_map_source(source, 1);
+	status = pthread_mutex_lock(&suc.mutex);
+	if (status) {
+		crit(ap->logopt,
+		     MODPREFIX "failed to lock startup condition mutex!");
+		cache_release(source);
 		master_free_mapent(entry);
 		return 1;
 	}
 
-	suc.ap = nap;
-	suc.root = mountpoint;
 	suc.done = 0;
 	suc.status = 0;
 
-	if (pthread_create(&thid, &th_attr_detached, handle_mounts, &suc)) {
+	if (pthread_create(&thid, NULL, handle_mounts, nap)) {
 		crit(ap->logopt,
 		     MODPREFIX
 		     "failed to create mount handler thread for %s",
-		     realpath);
-		handle_mounts_startup_cond_destroy(&suc);
+		     fullpath);
 		mounts_mutex_unlock(ap);
-		master_free_map_source(source, 1);
+		status = pthread_mutex_unlock(&suc.mutex);
+		if (status)
+			fatal(status);
+		cache_release(source);
 		master_free_mapent(entry);
 		return 1;
 	}
+	nap->thid = thid;
 
 	while (!suc.done) {
 		status = pthread_cond_wait(&suc.cond, &suc.mutex);
 		if (status) {
-			handle_mounts_startup_cond_destroy(&suc);
 			mounts_mutex_unlock(ap);
-			master_free_map_source(source, 1);
-			master_free_mapent(entry);
+			pthread_mutex_unlock(&suc.mutex);
 			fatal(status);
 		}
 	}
 
 	if (suc.status) {
 		crit(ap->logopt,
-		     MODPREFIX "failed to create submount for %s", realpath);
-		handle_mounts_startup_cond_destroy(&suc);
+		     MODPREFIX "failed to create submount for %s", fullpath);
 		mounts_mutex_unlock(ap);
-		master_free_map_source(source, 1);
+		status = pthread_mutex_unlock(&suc.mutex);
+		if (status)
+			fatal(status);
 		master_free_mapent(entry);
 		return 1;
 	}
-	nap->thid = thid;
 
 	ap->submnt_count++;
 	list_add(&nap->mounts, &ap->submounts);
 
-	handle_mounts_startup_cond_destroy(&suc);
 	mounts_mutex_unlock(ap);
 
+	status = pthread_mutex_unlock(&suc.mutex);
+	if (status)
+		fatal(status);
+
 	return 0;
 }
 
