@@ -15,6 +15,8 @@
 
 #include <stdio.h>
 #include <malloc.h>
+#include <fcntl.h>
+#include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
 #include <sys/param.h>
@@ -57,8 +59,7 @@ int mount_init(void **context)
 		bind_works = 1;
 	}
 
-	if (spawn_umount(LOGOPT_NONE, "-n", t2_dir, NULL) != 0)
-		debug(LOGOPT_ANY, MODPREFIX "umount failed for %s", t2_dir);
+	spawn_umount(LOGOPT_NONE, "-n", t2_dir, NULL);
 
 out:
 	rmdir(t1_dir);
@@ -70,78 +71,44 @@ out:
 int mount_mount(struct autofs_point *ap, const char *root, const char *name, int name_len,
 		const char *what, const char *fstype, const char *options, void *context)
 {
-	char fullpath[PATH_MAX];
+	char *fullpath;
 	char buf[MAX_ERR_BUF];
 	int err;
-	int i, len;
-	int symlnk = (*name != '/' && (ap->flags & MOUNT_FLAG_SYMLINK));
-
-	if (ap->flags & MOUNT_FLAG_REMOUNT)
-		return 0;
-
-	/* Extract "symlink" pseudo-option which forces local filesystems
-	 * to be symlinked instead of bound.
-	 */
-	if (*name != '/' && !symlnk && options) {
-		const char *comma;
-		int o_len = strlen(options) + 1;
-
-		for (comma = options; *comma != '\0';) {
-			const char *cp;
-			const char *end;
-
-			while (*comma == ',')
-				comma++;
-
-			/* Skip leading white space */
-			while (*comma == ' ' || *comma == '\t')
-				comma++;
-
-			cp = comma;
-			while (*comma != '\0' && *comma != ',')
-				comma++;
-
-			/* Skip trailing white space */
-			end = comma - 1;
-			while (*comma == ' ' || *comma == '\t')
-				end--;
-
-			o_len = end - cp + 1;
-			if (_strncmp("symlink", cp, o_len) == 0)
-				symlnk = 1;
-		}
-	}
+	int i, rlen;
 
 	/* Root offset of multi-mount */
-	len = strlen(root);
-	if (root[len - 1] == '/') {
-		len = snprintf(fullpath, len, "%s", root);
-	} else if (*name == '/') {
-		/*
-		 * Direct or offset mount, name is absolute path so
-		 * don't use root (but with move mount changes root
-		 * is now the same as name).
-		 */
-		len = sprintf(fullpath, "%s", root);
-	} else {
-		len = sprintf(fullpath, "%s/%s", root, name);
-	}
-	fullpath[len] = '\0';
+	if (*name == '/' && name_len == 1) {
+		rlen = strlen(root);
+		name_len = 0;
+	/* Direct mount name is absolute path so don't use root */
+	} else if (*name == '/')
+		rlen = 0;
+	else
+		rlen = strlen(root);
 
-	i = len;
+	fullpath = alloca(rlen + name_len + 2);
+	if (!fullpath) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "alloca: %s", estr);
+		return 1;
+	}
+
+	if (name_len) {
+		if (rlen)
+			sprintf(fullpath, "%s/%s", root, name);
+		else
+			sprintf(fullpath, "%s", name);
+	} else
+		sprintf(fullpath, "%s", root);
+
+	i = strlen(fullpath);
 	while (--i > 0 && fullpath[i] == '/')
 		fullpath[i] = '\0';
 
 	if (options == NULL || *options == '\0')
 		options = "defaults";
 
-	if (!strcmp(what, fullpath)) {
-		debug(ap->logopt, MODPREFIX
-		     "cannot mount or symlink %s to itself", fullpath);
-		return 1;
-	}
-
-	if (!symlnk && bind_works) {
+	if (bind_works) {
 		int status, existed = 1;
 
 		debug(ap->logopt, MODPREFIX "calling mkdir_path %s", fullpath);
@@ -158,19 +125,26 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 		if (!status)
 			existed = 0;
 
-		debug(ap->logopt, MODPREFIX
-		      "calling mount --bind -o %s %s %s",
+		if (is_mounted(_PATH_MOUNTED, fullpath, MNTS_REAL)) {
+			error(ap->logopt,
+			      MODPREFIX "warning: %s is already mounted",
+			      fullpath);
+			return 0;
+		}
+
+		debug(ap->logopt,
+		      MODPREFIX
+		      "calling mount --bind " SLOPPY " -o %s %s %s",
 		      options, what, fullpath);
 
-		err = spawn_bind_mount(ap->logopt, "-o",
-				       options, what, fullpath, NULL);
+		err = spawn_bind_mount(ap->logopt,
+			     SLOPPYOPT "-o", options, what, fullpath, NULL);
 
 		if (err) {
 			if (ap->type != LKP_INDIRECT)
 				return 1;
 
-			if (!existed &&
-			   (!(ap->flags & MOUNT_FLAG_GHOST) && name_len))
+			if ((!ap->ghost && name_len) || !existed)
 				rmdir_path(ap, fullpath, ap->dev);
 
 			return err;
@@ -182,7 +156,7 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 		}
 	} else {
 		char *cp;
-		char basepath[PATH_MAX];
+		char *basepath = alloca(strlen(fullpath) + 1);
 		int status;
 		struct stat st;
 
@@ -211,16 +185,11 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 		if (symlink(what, fullpath) && errno != EEXIST) {
 			error(ap->logopt,
 			      MODPREFIX
-			      "failed to create symlink %s -> %s",
+			      "failed to create local mount %s -> %s",
 			      fullpath, what);
-			if ((ap->flags & MOUNT_FLAG_GHOST) && !status) {
-				if (mkdir_path(fullpath, 0555) && errno != EEXIST) {
-					char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-					error(ap->logopt,
-					      MODPREFIX "mkdir_path %s failed: %s",
-					      fullpath, estr);
-				}
-			} else {
+			if (ap->ghost && !status)
+				mkdir_path(fullpath, 0555);
+			else {
 				if (ap->type == LKP_INDIRECT)
 					rmdir_path(ap, fullpath, ap->dev);
 			}
