@@ -19,6 +19,8 @@
 
 #include <stdio.h>
 #include <malloc.h>
+#include <fcntl.h>
+#include <unistd.h>
 #include <string.h>
 #include <sys/param.h>
 #include <sys/types.h>
@@ -41,39 +43,40 @@ int mount_init(void **context)
 	return 0;
 }
 
-int mount_reinit(void **context)
-{
-	return 0;
-}
-
 int mount_mount(struct autofs_point *ap, const char *root, const char *name, int name_len,
 		const char *what, const char *fstype, const char *options, void *context)
 {
-	char fullpath[PATH_MAX];
+	char *fullpath;
 	char buf[MAX_ERR_BUF];
 	int err;
-	int len, status, existed = 1;
-
-	if (ap->flags & MOUNT_FLAG_REMOUNT)
-		return 0;
+	int rlen, status, existed = 1;
 
 	fstype = "iso9660";
 
 	/* Root offset of multi-mount */
-	len = strlen(root);
-	if (root[len - 1] == '/') {
-		len = snprintf(fullpath, len, "%s", root);
-	} else if (*name == '/') {
-		/*
-		 * Direct or offset mount, name is absolute path so
-		 * don't use root (but with move mount changes root
-		 * is now the same as name).
-		 */
-		len = sprintf(fullpath, "%s", root);
-	} else {
-		len = sprintf(fullpath, "%s/%s", root, name);
+	if (*name == '/' && name_len == 1) {
+		rlen = strlen(root);
+		 name_len = 0;
+	/* Direct mount name is absolute path so don't use root */
+	} else if (*name == '/')
+		rlen = 0;
+	else
+		rlen = strlen(root);
+
+	fullpath = alloca(rlen + name_len + 2);
+	if (!fullpath) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "alloca: %s", estr);
+		return 1;
 	}
-	fullpath[len] = '\0';
+
+	if (name_len) {
+		if (rlen)
+			sprintf(fullpath, "%s/%s", root, name);
+		else
+			sprintf(fullpath, "%s", name);
+	} else
+		sprintf(fullpath, "%s", root);
 
 	debug(ap->logopt, MODPREFIX "calling umount %s", what);
 
@@ -108,12 +111,12 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 	}
 
 	if (options && options[0]) {
-		debug(ap->logopt, MODPREFIX
-		      "calling mount -t %s -o %s %s %s",
+		debug(ap->logopt,
+		      MODPREFIX "calling mount -t %s " SLOPPY "-o %s %s %s",
 		      fstype, options, what, fullpath);
 
 		err = spawn_mount(ap->logopt, "-t", fstype,
-			          "-o", options, what, fullpath, NULL);
+			     SLOPPYOPT "-o", options, what, fullpath, NULL);
 	} else {
 		debug(ap->logopt,
 		      MODPREFIX "calling mount -t %s %s %s",
@@ -129,12 +132,12 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 		if (ap->type != LKP_INDIRECT)
 			return 1;
 
-		if ((!(ap->flags & MOUNT_FLAG_GHOST) && name_len) || !existed)
+		if ((!ap->ghost && name_len) || !existed)
 			rmdir_path(ap, fullpath, ap->dev);
 
 		return 1;
 	} else {
-		debug(ap->logopt, MODPREFIX "mounted %s type %s on %s",
+		info(ap->logopt, MODPREFIX "mounted %s type %s on %s",
 		    what, fstype, fullpath);
 		return 0;
 	}
@@ -149,26 +152,31 @@ int swapCD(const char *device, const char *slotName)
 {
 	int fd;			/* file descriptor for CD-ROM device */
 	int status;		/* return status for system calls */
+	int cl_flags;
 	int slot = -1;
 	int total_slots_available;
 
 	slot = atoi(slotName) - 1;
 
 	/* open device */
-	fd = open_fd(device, O_RDONLY | O_NONBLOCK);
+	fd = open(device, O_RDONLY | O_NONBLOCK);
 	if (fd < 0) {
 		logerr(MODPREFIX "Opening device %s failed : %s",
 		      device, strerror(errno));
 		return 1;
 	}
 
+	if ((cl_flags = fcntl(fd, F_GETFD, 0)) != -1) {
+		cl_flags |= FD_CLOEXEC;
+		fcntl(fd, F_SETFD, cl_flags);
+	}
+
 	/* Check CD player status */
 	total_slots_available = ioctl(fd, CDROM_CHANGER_NSLOTS);
 	if (total_slots_available <= 1) {
 		logerr(MODPREFIX
 		      "Device %s is not an ATAPI compliant CD changer.",
 		      device);
-		close(fd);
 		return 1;
 	}
 
@@ -176,7 +184,6 @@ int swapCD(const char *device, const char *slotName)
 	slot = ioctl(fd, CDROM_SELECT_DISC, slot);
 	if (slot < 0) {
 		logerr(MODPREFIX "CDROM_SELECT_DISC failed");
-		close(fd);
 		return 1;
 	}
 
