@@ -15,6 +15,8 @@
 
 #include <stdio.h>
 #include <malloc.h>
+#include <fcntl.h>
+#include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
 #include <sys/param.h>
@@ -33,47 +35,44 @@ int mount_init(void **context)
 	return 0;
 }
 
-int mount_reinit(void **context)
-{
-	return 0;
-}
-
 int mount_mount(struct autofs_point *ap, const char *root, const char *name, int name_len,
 		const char *what, const char *fstype, const char *options, void *context)
 {
-	char fullpath[PATH_MAX];
+	char *fullpath;
 	char buf[MAX_ERR_BUF];
 	const char *p, *p1;
 	int err, ro = 0;
 	const char *fsck_prog;
-	int len, status, existed = 1;
-	void (*mountlog)(unsigned int, const char*, ...) = &log_debug;
-
-	if (ap->flags & MOUNT_FLAG_REMOUNT)
-		return 0;
-
-	if (defaults_get_mount_verbose())
-		mountlog = &log_info;
+	int rlen, status, existed = 1;
 
 	/* Root offset of multi-mount */
-	len = strlen(root);
-	if (root[len - 1] == '/') {
-		len = snprintf(fullpath, len, "%s", root);
-	} else if (*name == '/') {
-		/*
-		 * Direct or offset mount, name is absolute path so
-		 * don't use root (but with move mount changes root
-		 * is now the same as name).
-		 */
-		len = sprintf(fullpath, "%s", root);
-	} else {
-		len = sprintf(fullpath, "%s/%s", root, name);
+	if (*name == '/' && name_len == 1) {
+		rlen = strlen(root);
+		name_len = 0;
+	/* Direct mount name is absolute path so don't use root */
+	} else if (*name == '/')
+		rlen = 0;
+	else
+		rlen = strlen(root);
+
+	fullpath = alloca(rlen + name_len + 2);
+	if (!fullpath) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "alloca: %s", estr);
+		return 1;
 	}
-	fullpath[len] = '\0';
+
+	if (name_len) {
+		if (rlen)
+			sprintf(fullpath, "%s/%s", root, name);
+		else
+			sprintf(fullpath, "%s", name);
+	} else
+		sprintf(fullpath, "%s", root);
 
 	debug(ap->logopt, MODPREFIX "calling mkdir_path %s", fullpath);
 
-	status = mkdir_path(fullpath, mp_mode);
+	status = mkdir_path(fullpath, 0555);
 	if (status && errno != EEXIST) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		error(ap->logopt,
@@ -84,22 +83,27 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 	if (!status)
 		existed = 0;
 
+	if (is_mounted(_PATH_MOUNTED, fullpath, MNTS_REAL)) {
+		error(ap->logopt,
+		      MODPREFIX "warning: %s is already mounted", fullpath);
+		return 0;
+	}
+
 	if (options && options[0]) {
 		for (p = options; (p1 = strchr(p, ',')); p = p1)
-			if (!_strncmp("ro", p, p1 - p) && ++p1 - p == sizeof("ro"))
+			if (!strncmp(p, "ro", p1 - p) && ++p1 - p == sizeof("ro"))
 				ro = 1;
 		if (!strcmp(p, "ro"))
 			ro = 1;
 	}
 
-	fsck_prog = PATH_E2FSCK;
 #ifdef HAVE_E3FSCK
-	if (!strcmp(fstype,"ext3"))
+	if (!strcmp(fstype,"ext3") || !strcmp(fstype,"auto"))
 		fsck_prog = PATH_E3FSCK;
-#endif
-#ifdef HAVE_E4FSCK
-	if (!strcmp(fstype,"ext4"))
-		fsck_prog = PATH_E4FSCK;
+	else
+		fsck_prog = PATH_E2FSCK;
+#else
+	fsck_prog = PATH_E2FSCK;
 #endif
 	if (ro) {
 		debug(ap->logopt,
@@ -125,15 +129,15 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 	}
 
 	if (options) {
-		mountlog(ap->logopt, MODPREFIX
-			 "calling mount -t %s -o %s %s %s",
-			 fstype, options, what, fullpath);
+		debug(ap->logopt,
+		      MODPREFIX "calling mount -t %s " SLOPPY "-o %s %s %s",
+		      fstype, options, what, fullpath);
 		err = spawn_mount(ap->logopt, "-t", fstype,
-			          "-o", options, what, fullpath, NULL);
+			     SLOPPYOPT "-o", options, what, fullpath, NULL);
 	} else {
-		mountlog(ap->logopt,
-			 MODPREFIX "calling mount -t %s %s %s",
-			 fstype, what, fullpath);
+		debug(ap->logopt,
+		      MODPREFIX "calling mount -t %s %s %s",
+		      fstype, what, fullpath);
 		err = spawn_mount(ap->logopt, "-t", fstype, what, fullpath, NULL);
 	}
 
@@ -144,12 +148,12 @@ int mount_mount(struct autofs_point *ap, const char *root, const char *name, int
 		if (ap->type != LKP_INDIRECT)
 			return 1;
 
-		if ((!(ap->flags & MOUNT_FLAG_GHOST) && name_len) || !existed)
+		if ((!ap->ghost && name_len) || !existed)
 			rmdir_path(ap, fullpath, ap->dev);
 
 		return 1;
 	} else {
-		debug(ap->logopt,
+		info(ap->logopt,
 		      MODPREFIX "mounted %s type %s on %s",
 		      what, fstype, fullpath);
 		return 0;
