@@ -7,6 +7,7 @@
 #include <sys/types.h>
 #include <ctype.h>
 #include <string.h>
+#include <unistd.h>
 #include <stdlib.h>
 #include <netinet/in.h>
 #include <arpa/nameser.h>
@@ -46,12 +47,6 @@ static int parse_afs(struct autofs_point *ap,
 
 	/* Isolate the source for this AFS fs. */
 	for (i = 0; (!isspace(p[i]) && i < source_len); i++) {
-		if (!p[i]) {
-			error(ap->logopt, MODPREFIX
-			      "unexpeced end of input looking for AFS "
-			      "source: %s", p);
-			return 1;
-		}
 		source[i] = p[i];
 	}
 
@@ -62,14 +57,8 @@ static int parse_afs(struct autofs_point *ap,
 	while ((*p) && (isspace(*p)))
 		p++;
 
-	/* Isolate the options for this AFS fs. */
+	/* Isolate the source for this AFS fs. */
 	for (i = 0; (!isspace(p[i]) && i < options_len); i++) {
-		if (!p[i]) {
-			error(ap->logopt, MODPREFIX
-			      "unexpeced end of input looking for AFS "
-			      "options: %s", p);
-			return 1;
-		}
 		options[i] = p[i];
 	}
 	options[i] = 0;
@@ -117,13 +106,7 @@ static int parse_nfs(struct autofs_point *ap,
 		p++;
 
 	/* Isolate the remote mountpoint for this NFS fs. */
-	for (i = 0; (!isspace(p[i]) && i < ((int) sizeof(mount) - 1)); i++) {
-		if (!p[i]) {
-			error(ap->logopt, MODPREFIX
-			      "unexpeced end of input looking for NFS "
-			      "mountpoint: %s", p);
-			return 1;
-		}
+	for (i = 0; (!isspace(p[i]) && i < (int) sizeof(mount)); i++) {
 		mount[i] = p[i];
 	}
 
@@ -136,26 +119,15 @@ static int parse_nfs(struct autofs_point *ap,
 
 	/* Isolate the remote host. */
 	for (i = 0; (!isspace(p[i]) && i < source_len); i++) {
-		if (!p[i]) {
-			error(ap->logopt, MODPREFIX
-			      "unexpeced end of input looking for NFS "
-			      "host: %s", p);
-			return 1;
-		}
 		source[i] = p[i];
 	}
 
 	source[i] = 0;
 	p += i;
 
-	if (strlen(source) + strlen(mount) + 2 > source_len) {
-		error(ap->logopt, MODPREFIX "entry too log for mount source");
-		return 1;
-	}
-
 	/* Append ":mountpoint" to the source to get "host:mountpoint". */
-	strcat(source, ":");
-	strcat(source, mount);
+	strncat(source, ":", source_len);
+	strncat(source, mount, source_len);
 
 	/* Skip whitespace. */
 	while ((*p) && (isspace(*p)))
@@ -163,12 +135,6 @@ static int parse_nfs(struct autofs_point *ap,
 
 	/* Isolate the mount options. */
 	for (i = 0; (!isspace(p[i]) && i < options_len); i++) {
-		if (!p[i]) {
-			error(ap->logopt, MODPREFIX
-			      "unexpeced end of input looking for NFS "
-			      "mount options: %s", p);
-			return 1;
-		}
 		options[i] = p[i];
 	}
 	options[i] = 0;
@@ -213,12 +179,6 @@ static int parse_generic(struct autofs_point *ap,
 
 	/* Isolate the source for this fs. */
 	for (i = 0; (!isspace(p[i]) && i < source_len); i++) {
-		if (!p[i]) {
-			error(ap->logopt, MODPREFIX
-			      "unexpeced end of input looking for generic "
-			      "mount source: %s", p);
-			return 1;
-		}
 		source[i] = p[i];
 	}
 
@@ -231,12 +191,6 @@ static int parse_generic(struct autofs_point *ap,
 
 	/* Isolate the mount options. */
 	for (i = 0; (!isspace(p[i]) && i < options_len); i++) {
-		if (!p[i]) {
-			error(ap->logopt, MODPREFIX
-			      "unexpeced end of input looking for generic "
-			      "mount options: %s", p);
-			return 1;
-		}
 		options[i] = p[i];
 	}
 	options[i] = 0;
@@ -274,7 +228,6 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	char options[HESIOD_LEN + 1];
 	char *q;
 	const char *p;
-	int ret;
 
 	ap->entry->current = NULL;
 	master_source_current_signal(ap->entry);
@@ -298,28 +251,19 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		return 1;
 	/* If it's an AFS fs... */
 	} else if (!strcasecmp(fstype, "afs"))
-		ret = parse_afs(ap, mapent, name, name_len,
-				source, sizeof(source), options,
-				sizeof(options));
+		parse_afs(ap, mapent, name, name_len,
+			  source, sizeof(source), options, sizeof(options));
 	/* If it's NFS... */
 	else if (!strcasecmp(fstype, "nfs"))
-		ret = parse_nfs(ap, mapent, name, name_len,
-				source, sizeof(source), options,
-				sizeof(options));
+		parse_nfs(ap, mapent, name, name_len,
+			  source, sizeof(source), options, sizeof(options));
 	/* Punt. */
 	else
-		ret = parse_generic(ap, mapent, name, name_len,
-				    source, sizeof(source), options,
-				    sizeof(options));
+		parse_generic(ap, mapent, name, name_len, source, sizeof(source),
+			      options, sizeof(options));
 
-	if (ret) {
-		error(ap->logopt, MODPREFIX "failed to parse entry");
-		return 1;
-	} else {
-		debug(ap->logopt,
-		      MODPREFIX "mount %s is type %s from %s",
-		      name, fstype, source);
-	}
+	debug(ap->logopt,
+	      MODPREFIX "mount %s is type %s from %s", name, fstype, source);
 
 	return do_mount(ap, ap->path, name, name_len, source, fstype, options);
 }
