@@ -18,6 +18,8 @@
 #include <stdio.h>
 #include <malloc.h>
 #include <netdb.h>
+#include <fcntl.h>
+#include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -29,8 +31,6 @@
 #include <sys/vfs.h>
 #include <sys/utsname.h>
 #include <netinet/in.h>
-#include <sys/mount.h>
-#include <linux/fs.h>
 
 #define MODULE_PARSE
 #include "automount.h"
@@ -41,21 +41,6 @@ int parse_version = AUTOFS_PARSE_VERSION;	/* Required by protocol */
 
 static struct mount_mod *mount_nfs = NULL;
 static int init_ctr = 0;
-static pthread_mutex_t instance_mutex = PTHREAD_MUTEX_INITIALIZER;
-
-static void instance_mutex_lock(void)
-{
-	int status = pthread_mutex_lock(&instance_mutex);
-	if (status)
-		fatal(status);
-}
-
-static void instance_mutex_unlock(void)
-{
-	int status = pthread_mutex_unlock(&instance_mutex);
-	if (status)
-		fatal(status);
-}
 
 extern const char *global_options;
 
@@ -82,7 +67,6 @@ static struct parse_context default_context = {
 	1			/* Do slashify_colons */
 };
 
-int destroy_logpri_fifo(struct autofs_point *ap);
 static char *concat_options(char *left, char *right);
 
 /* Free all storage associated with this context */
@@ -98,6 +82,44 @@ static void kill_context(struct parse_context *ctxt)
 	free(ctxt);
 }
 
+static struct substvar *addstdenv(struct substvar *sv)
+{
+	struct substvar *list = sv;
+	struct thread_stdenv_vars *tsv;
+	char numbuf[16];
+
+	tsv = pthread_getspecific(key_thread_stdenv_vars);
+	if (tsv) {
+		int ret;
+		long num;
+
+		num = (long) tsv->uid;
+		ret = sprintf(numbuf, "%ld", num);
+		if (ret > 0)
+			list = macro_addvar(list, "UID", 3, numbuf);
+		num = (long) tsv->gid;
+		ret = sprintf(numbuf, "%ld", num);
+		if (ret > 0)
+			list = macro_addvar(list, "GID", 3, numbuf);
+		list = macro_addvar(list, "USER", 4, tsv->user);
+		list = macro_addvar(list, "GROUP", 5, tsv->group);
+		list = macro_addvar(list, "HOME", 4, tsv->home);
+	}
+	return list;
+}
+
+static struct substvar *removestdenv(struct substvar *sv)
+{
+	struct substvar *list = sv;
+
+	list = macro_removevar(list, "UID", 3);
+	list = macro_removevar(list, "USER", 4);
+	list = macro_removevar(list, "HOME", 4);
+	list = macro_removevar(list, "GID", 3);
+	list = macro_removevar(list, "GROUP", 5);
+	return list;
+}
+
 /* 
  * $- and &-expand a Sun-style map entry and return the length of the entry.
  * If "dst" is NULL, just count the length.
@@ -118,24 +140,30 @@ int expandsunent(const char *src, char *dst, const char *key,
 		case '&':
 			l = strlen(key);
 			/*
-			 * In order to ensure that any isspace() characters
-			 * in the key are preserved, we need to escape them
-			 * here.
+			 * In order to ensure that any spaces in the key
+			 * re preserved, we need to escape them here.
 			 */
-			const char *keyp = key;
-			while (*keyp) {
-				if (isspace(*keyp)) {
-					if (dst) {
-						*dst++ = '\\';
-						*dst++ = *keyp++;
-					} else
-						keyp++;
-					l++;
-				} else {
-					if (dst)
-						*dst++ = *keyp++;
-					else
-						keyp++;
+			if (strchr(key, ' ')) {
+				const char *keyp = key;
+				while (*keyp) {
+					if (isspace(*keyp)) {
+						if (dst) {
+							*dst++ = '\\';
+							*dst++ = *keyp++;
+						} else
+							keyp++;
+						l++;
+					} else {
+						if (dst)
+							*dst++ = *keyp++;
+						else
+							keyp++;
+					}
+				}
+			} else {
+				if (dst) {
+					strcpy(dst, key);
+					dst += l;
 				}
 			}
 			len += l;
@@ -212,9 +240,7 @@ int expandsunent(const char *src, char *dst, const char *key,
 				*(dst++) = 
 				  (seen_colons && slashify_colons) ? '/' : ':';
 			len++;
-			/* Were looking for the colon preceeding a path */
-			if (*src == '/')
-				seen_colons = 1;
+			seen_colons = 1;
 			break;
 
 		default:
@@ -232,15 +258,32 @@ int expandsunent(const char *src, char *dst, const char *key,
 	return len;
 }
 
-static int do_init(int argc, const char *const *argv, struct parse_context *ctxt)
+int parse_init(int argc, const char *const *argv, void **context)
 {
-	char *noptstr, *def, *val, *macros, *gbl_options;
+	struct parse_context *ctxt;
 	char buf[MAX_ERR_BUF];
-	int optlen, len, offset;
+	char *noptstr, *def, *val, *macros, *gbl_options;
 	const char *xopt;
+	int optlen, len, offset;
 	int i, bval;
 	unsigned int append_options;
 
+	/* Get processor information for predefined escapes */
+
+	if (!init_ctr)
+		macro_init();
+
+	/* Set up context and escape chain */
+
+	if (!(ctxt = (struct parse_context *) malloc(sizeof(struct parse_context)))) {
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "malloc: %s", estr);
+		*context = NULL;
+		return 1;
+	}
+	*context = (void *) ctxt;
+
+	*ctxt = default_context;
 	optlen = 0;
 
 	/* Look for options and capture, and create new defines if we need to */
@@ -331,21 +374,21 @@ static int do_init(int argc, const char *const *argv, struct parse_context *ctxt
 			if (ctxt->optstr) {
 				noptstr =
 				    (char *) realloc(ctxt->optstr, optlen + len + 2);
-				if (noptstr) {
-					noptstr[optlen] = ',';
-					strcpy(noptstr + optlen + 1, argv[i] + offset);
-					optlen += len + 1;
-				}
+				if (!noptstr)
+					break;
+				noptstr[optlen] = ',';
+				strcpy(noptstr + optlen + 1, argv[i] + offset);
+				optlen += len + 1;
 			} else {
 				noptstr = (char *) malloc(len + 1);
-				if (noptstr) {
-					strcpy(noptstr, argv[i] + offset);
-					optlen = len;
-				}
+				strcpy(noptstr, argv[i] + offset);
+				optlen = len;
 			}
 			if (!noptstr) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				kill_context(ctxt);
 				logerr(MODPREFIX "%s", estr);
+				*context = NULL;
 				return 1;
 			}
 			ctxt->optstr = noptstr;
@@ -377,80 +420,24 @@ static int do_init(int argc, const char *const *argv, struct parse_context *ctxt
 		}
 	}
 options_done:
-
 	debug(LOGOPT_NONE,
 	      MODPREFIX "init gathered global options: %s", ctxt->optstr);
 
-	return 0;
-}
-
-int parse_init(int argc, const char *const *argv, void **context)
-{
-	struct parse_context *ctxt;
-	char buf[MAX_ERR_BUF];
-
-	*context = NULL;
-
-	/* Set up context and escape chain */
-
-	ctxt = (struct parse_context *) malloc(sizeof(struct parse_context));
-	if (!ctxt) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		return 1;
-	}
-
-	*ctxt = default_context;
-
-	if (do_init(argc, argv, ctxt)) {
-		free(ctxt);
-		return 1;
-	}
-
 	/* We only need this once.  NFS mounts are so common that we cache
 	   this module. */
-	instance_mutex_lock();
-	if (mount_nfs)
-		init_ctr++;
-	else {
+	if (!mount_nfs) {
 		if ((mount_nfs = open_mount("nfs", MODPREFIX))) {
 			init_ctr++;
+			return 0;
 		} else {
 			kill_context(ctxt);
-			instance_mutex_unlock();
+			*context = NULL;
 			return 1;
 		}
+	} else {
+		init_ctr++;
+		return 0;
 	}
-	instance_mutex_unlock();
-
-	*context = (void *) ctxt;
-
-	return 0;
-}
-
-int parse_reinit(int argc, const char *const *argv, void **context)
-{
-	struct parse_context *ctxt = (struct parse_context *) *context;
-	struct parse_context *new;
-	char buf[MAX_ERR_BUF];
-
-	new = (struct parse_context *) malloc(sizeof(struct parse_context));
-	if (!new) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr(MODPREFIX "malloc: %s", estr);
-		return 1;
-	}
-
-	*new = default_context;
-
-	if (do_init(argc, argv, new))
-		return 1;
-
-	kill_context(ctxt);
-
-	*context = (void *) new;
-
-	return 0;
 }
 
 static const char *parse_options(const char *str, char **ret, unsigned int logopt)
@@ -475,17 +462,11 @@ static char *concat_options(char *left, char *right)
 	char buf[MAX_ERR_BUF];
 	char *ret;
 
-	if (left == NULL || *left == '\0') {
-		ret = strdup(right);
-		free(right);
-		return ret;
-	}
+	if (left == NULL || *left == '\0')
+		return strdup(right);
 
-	if (right == NULL || *right == '\0') {
-		ret = strdup(left);
-		free(left);
-		return ret;
-	}
+	if (right == NULL || *right == '\0')
+		return strdup(left);
 
 	ret = malloc(strlen(left) + strlen(right) + 2);
 
@@ -512,7 +493,6 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 {
 	char *fstype = "nfs";	/* Default filesystem type */
 	int nonstrict = 1;
-	int use_weight_only = ap->flags & MOUNT_FLAG_USE_WEIGHT_ONLY;
 	int rv, cur_state;
 	char *mountpoint;
 	char *what;
@@ -542,30 +522,25 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 			while (*comma != '\0' && *comma != ',')
 				comma++;
 
-			if (_strncmp("fstype=", cp, 7) == 0) {
+			if (strncmp("fstype=", cp, 7) == 0) {
 				int typelen = comma - (cp + 7);
 				fstype = alloca(typelen + 1);
 				memcpy(fstype, cp + 7, typelen);
 				fstype[typelen] = '\0';
-			} else if (_strncmp("nonstrict", cp, 9) == 0) {
+			} else if (strncmp("nonstrict", cp, 9) == 0) {
 				nonstrict = 1;
-			} else if (_strncmp("strict", cp, 6) == 0 &&
-				   comma - cp == 6) {
+			} else if (strncmp("strict", cp, 6) == 0) {
 				nonstrict = 0;
-			} else if (_strncmp("nobrowse", cp, 8) == 0 ||
-				   _strncmp("browse", cp, 6) == 0 ||
-				   _strncmp("timeout=", cp, 8) == 0) {
+			} else if (strncmp("nobrowse", cp, 8) == 0 ||
+				   strncmp("browse", cp, 6) == 0 ||
+				   strncmp("timeout=", cp, 8) == 0) {
 				if (strcmp(fstype, "autofs") == 0 ||
 				    strstr(cp, "fstype=autofs")) {
 					memcpy(np, cp, comma - cp + 1);
 					np += comma - cp + 1;
 				}
-			} else if (_strncmp("no-use-weight-only", cp, 18) == 0) {
-				use_weight_only = -1;
-			} else if (_strncmp("use-weight-only", cp, 15) == 0) {
-				use_weight_only = MOUNT_FLAG_USE_WEIGHT_ONLY;
-			} else if (_strncmp("bg", cp, 2) == 0 ||
-				   _strncmp("nofg", cp, 4) == 0) {
+			} else if (strncmp("bg", cp, 2) == 0 ||
+				   strncmp("nofg", cp, 4) == 0) {
 				continue;
 			} else {
 				memcpy(np, cp, comma - cp + 1);
@@ -582,10 +557,11 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		options = noptions;
 	}
 
+
 	if (!strcmp(fstype, "autofs") && ctxt->macros) {
 		char *noptions = NULL;
 
-		if (!options || *options == '\0') {
+		if (!options) {
 			noptions = alloca(strlen(ctxt->macros) + 1);
 			*noptions = '\0';
 		} else {
@@ -598,7 +574,7 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 			}
 		}
 
-		if (noptions && *noptions != '\0') {
+		if (noptions) {
 			strcat(noptions, ctxt->macros);
 			options = noptions;
 		} else {
@@ -612,14 +588,13 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 
 	type = ap->entry->maps->type;
 	if (type && !strcmp(type, "hosts")) {
-		if (options && *options != '\0') {
+		if (options) {
 			int len = strlen(options);
 			int suid = strstr(options, "suid") ? 0 : 7;
 			int dev = strstr(options, "dev") ? 0 : 6;
-			int nointr = strstr(options, "nointr") ? 0 : 5;
 
-			if (suid || dev || nointr) {
-				char *tmp = alloca(len + suid + dev + nointr + 1);
+			if (suid || dev) {
+				char *tmp = alloca(len + suid + dev + 1);
 				if (!tmp) {
 					error(ap->logopt, MODPREFIX
 					      "alloca failed for options");
@@ -633,12 +608,10 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 					strcat(tmp, ",nosuid");
 				if (dev)
 					strcat(tmp, ",nodev");
-				if (nointr)
-					strcat(tmp, ",intr");
 				options = tmp;
 			}
 		} else {
-			char *tmp = alloca(18);
+			char *tmp = alloca(13);
 			if (!tmp) {
 				error(ap->logopt,
 				      MODPREFIX "alloca failed for options");
@@ -646,41 +619,17 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 					return -1;
 				return 1;
 			}
-			strcpy(tmp, "nosuid,nodev,intr");
+			strcpy(tmp, "nosuid,nodev");
 			options = tmp;
 		}
 	}
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-	if (!strcmp(fstype, "nfs") || !strcmp(fstype, "nfs4")) {
+	if (!strcmp(fstype, "nfs")) {
 		what = alloca(loclen + 1);
 		memcpy(what, loc, loclen);
 		what[loclen] = '\0';
 
-		/* Add back "[no-]use-weight-only" for NFS mounts only */
-		if (use_weight_only) {
-			char *tmp;
-			int len;
-
-			if (options && *options != '\0') {
-				len = strlen(options) + 19;
-				tmp = alloca(len);
-				strcpy(tmp, options);
-				strcat(tmp, ",");
-				if (use_weight_only == MOUNT_FLAG_USE_WEIGHT_ONLY)
-					strcat(tmp, "use-weight-only");
-				else
-					strcat(tmp, "no-use-weight-only");
-			} else {
-				tmp = alloca(19);
-				if (use_weight_only == MOUNT_FLAG_USE_WEIGHT_ONLY)
-					strcpy(tmp, "use-weight-only");
-				else
-					strcpy(tmp, "no-use-weight-only");
-			}
-			options = tmp;
-		}
-
 		debug(ap->logopt, MODPREFIX
 		      "mounting root %s, mountpoint %s, "
 		      "what %s, fstype %s, options %s",
@@ -689,18 +638,14 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		rv = mount_nfs->mount_mount(ap, root, mountpoint, strlen(mountpoint),
 					    what, fstype, options, mount_nfs->context);
 	} else {
-		if (!loclen)
-			what = NULL;
-		else {
-			what = alloca(loclen + 1);
-			if (*loc == ':') {
-				loclen--;
-				memcpy(what, loc + 1, loclen);
-				what[loclen] = '\0';
-			} else {
-				memcpy(what, loc, loclen);
-				what[loclen] = '\0';
-			}
+		what = alloca(loclen + 1);
+		if (*loc == ':') {
+			loclen--;
+			memcpy(what, loc + 1, loclen);
+			what[loclen] = '\0';
+		} else {
+			memcpy(what, loc, loclen);
+			what[loclen] = '\0';
 		}
 
 		debug(ap->logopt, MODPREFIX
@@ -779,10 +724,10 @@ static int check_is_multi(const char *mapent)
 }
 
 static int
-update_offset_entry(struct autofs_point *ap, const char *name,
-		    const char *m_root, int m_root_len,
-		    const char *path, const char *myoptions, const char *loc,
-		    time_t age)
+add_offset_entry(struct autofs_point *ap, const char *name,
+		 const char *m_root, int m_root_len,
+		 const char *path, const char *myoptions, const char *loc,
+		 time_t age)
 {
 	struct map_source *source;
 	struct mapent_cache *mc;
@@ -797,10 +742,7 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 
 	mc = source->mc;
 
-	memset(m_mapent, 0, MAPENT_MAX_LEN + 1);
-
-	/* Internal hosts map may have loc == NULL */
-	if (!*path) {
+	if (!*path || !*loc) {
 		error(ap->logopt,
 		      MODPREFIX "syntax error in offset %s -> %s", path, loc);
 		return CHE_FAIL;
@@ -808,10 +750,8 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 
 	p_len = strlen(path);
 	/* Trailing '/' causes us pain */
-	if (p_len > 1) {
-		while (p_len > 1 && path[p_len - 1] == '/')
-			p_len--;
-	}
+	if (p_len > 1 && path[p_len - 1] == '/')
+		p_len--;
 	m_key_len = m_root_len + p_len;
 	if (m_key_len > PATH_MAX) {
 		error(ap->logopt, MODPREFIX "multi mount key too long");
@@ -825,7 +765,7 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 	if (*myoptions)
 		m_options_len = strlen(myoptions) + 2;
 
-	m_mapent_len = loc ? strlen(loc) : 0;
+	m_mapent_len = strlen(loc);
 	if (m_mapent_len + m_options_len > MAPENT_MAX_LEN) {
 		error(ap->logopt, MODPREFIX "multi mount mapent too long");
 		return CHE_FAIL;
@@ -834,34 +774,23 @@ update_offset_entry(struct autofs_point *ap, const char *name,
 	if (*myoptions) {
 		strcpy(m_mapent, "-");
 		strcat(m_mapent, myoptions);
-		if (loc) {
-			strcat(m_mapent, " ");
-			if (loc)
-				strcat(m_mapent, loc);
-		}
-	} else {
-		if (loc)
-			strcpy(m_mapent, loc);
-	}
+		strcat(m_mapent, " ");
+		strcat(m_mapent, loc);
+	} else
+		strcpy(m_mapent, loc);
 
-	ret = cache_update_offset(mc, name, m_key, m_mapent, age);
-	if (ret == CHE_DUPLICATE) {
+	ret = cache_add_offset(mc, name, m_key, m_mapent, age);
+	if (ret == CHE_OK)
+		debug(ap->logopt, MODPREFIX
+		      "added multi-mount offset %s -> %s", path, m_mapent);
+	else
 		warn(ap->logopt, MODPREFIX
-		     "syntax error or duplicate offset %s -> %s", path, loc);
-		ret = CHE_OK;
-	} else if (ret == CHE_FAIL)
-		debug(ap->logopt, MODPREFIX
-		      "failed to update multi-mount offset %s -> %s", path, m_mapent);
-	else {
-		ret = CHE_OK;
-		debug(ap->logopt, MODPREFIX
-		      "updated multi-mount offset %s -> %s", path, m_mapent);
-	}
+		      "syntax error or duplicate offset %s -> %s", path, loc);
 
 	return ret;
 }
 
-static int validate_location(unsigned int logopt, char *loc)
+static int validate_location(char *loc)
 {
 	char *ptr = loc;
 
@@ -870,53 +799,27 @@ static int validate_location(unsigned int logopt, char *loc)
 		return 1;
 
 	/*
-	 * If a ':/' is present now it must be a host name, except
+	 * If a ':' is present now it must be a host name, except
 	 * for those special file systems like sshfs which use "#"
-	 * and "@" in the host name part and ipv6 addresses that
-	 * have ":", "[" and "]".
+	 * and "@" in the host name part.
 	 */
-	if (!check_colon(ptr)) {
-		char *esc;
-		/*
-		 * Don't forget cases where a colon is present but
-		 * not followed by a "/" or, if there is no colon at
-		 * all, we don't know if it is actually invalid since
-		 * it may be a map name by itself, for example.
-		 */
-		if (!strchr(ptr, ':') ||
-		    ((esc = strchr(ptr, '\\')) && *(esc + 1) == ':') ||
-		    !strncmp(ptr, "file:", 5) || !strncmp(ptr, "yp:", 3) ||
-		    !strncmp(ptr, "nis:", 4) || !strncmp(ptr, "nisplus:", 8) ||
-		    !strncmp(ptr, "ldap:", 5) || !strncmp(ptr, "ldaps:", 6) ||
-		    !strncmp(ptr, "sss:", 4) || !strncmp(ptr, "dir:", 4))
-			return 1;
-		error(logopt,
-		      "expected colon delimeter not found in location %s",
-		      loc);
-		return 0;
-	} else {
-		while (*ptr && strncmp(ptr, ":/", 2)) {
+	if (check_colon(ptr)) {
+		while (*ptr && *ptr != ':') {
 			if (!(isalnum(*ptr) ||
 			    *ptr == '-' || *ptr == '.' || *ptr == '_' ||
 			    *ptr == ',' || *ptr == '(' || *ptr == ')' ||
-			    *ptr == '#' || *ptr == '@' || *ptr == ':' ||
-			    *ptr == '[' || *ptr == ']' || *ptr == '%')) {
-				error(logopt, "invalid character \"%c\" "
-				      "found in location %s", *ptr, loc);
+			    *ptr == '#' || *ptr == '@'))
 				return 0;
-			}
 			ptr++;
 		}
 
-		if (*ptr && !strncmp(ptr, ":/", 2))
+		if (*ptr && *ptr == ':')
 			ptr++;
 	}
 
 	/* Must always be something following */
-	if (!*ptr) {
-		error(logopt, "invalid location %s", loc);
+	if (!*ptr)
 		return 0;
-	}
 
 	return 1;
 }
@@ -970,26 +873,20 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	l = chunklen(p, check_colon(p));
 	loc = dequote(p, l, logopt);
 	if (!loc) {
-		if (strstr(myoptions, "fstype=autofs") &&
-		    strstr(myoptions, "hosts")) {
-			warn(logopt, MODPREFIX "possible missing location");
-			free(myoptions);
-			return 0;
-		}
-		*options = myoptions;
-		*location = NULL;
-		return (p - ent);
+		warn(logopt, MODPREFIX "possible missing location");
+		free(myoptions);
+		return 0;
 	}
 
 	/* Location can't begin with a '/' */
 	if (*p == '/') {
 		warn(logopt, MODPREFIX "error location begins with \"/\"");
 		free(myoptions);
-		free(loc);
 		return 0;
 	}
 
-	if (!validate_location(logopt, loc)) {
+	if (!validate_location(loc)) {
+		warn(logopt, MODPREFIX "invalid location %s", loc);
 		free(myoptions);
 		free(loc);
 		return 0;
@@ -1006,15 +903,10 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		l = chunklen(p, check_colon(p));
 		ent_chunk = dequote(p, l, logopt);
 		if (!ent_chunk) {
-			if (strstr(myoptions, "fstype=autofs") &&
-			    strstr(myoptions, "hosts")) {
-				warn(logopt, MODPREFIX
-				     "null location or out of memory");
-				free(myoptions);
-				free(loc);
-				return 0;
-			}
-			goto next;
+			warn(logopt, MODPREFIX "null location or out of memory");
+			free(myoptions);
+			free(loc);
+			return 0;
 		}
 
 		/* Location can't begin with a '/' */
@@ -1027,7 +919,9 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 			return 0;
 		}
 
-		if (!validate_location(logopt, ent_chunk)) {
+		if (!validate_location(ent_chunk)) {
+			warn(logopt,
+			      MODPREFIX "invalid location %s", ent_chunk);
 			free(ent_chunk);
 			free(myoptions);
 			free(loc);
@@ -1050,7 +944,7 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		strcat(loc, ent_chunk);
 
 		free(ent_chunk);
-next:
+
 		p += l;
 		p = skipspace(p);
 	}
@@ -1061,200 +955,53 @@ next:
 	return (p - ent);
 }
 
-static void cleanup_multi_triggers(struct autofs_point *ap,
-			    struct mapent *me, const char *root, int start,
-			    const char *base)
-{
-	char path[PATH_MAX + 1];
-	char offset[PATH_MAX + 1];
-	char *poffset = offset;
-	struct mapent *oe;
-	struct list_head *mm_root, *pos;
-	const char o_root[] = "/";
-	const char *mm_base;
-
-	mm_root = &me->multi->multi_list;
-
-	if (!base)
-		mm_base = o_root;
-	else
-		mm_base = base;
-
-	pos = NULL;
-
-	/* Make sure "none" of the offsets have an active mount. */
-	while ((poffset = cache_get_offset(mm_base, poffset, start, mm_root, &pos))) {
-		oe = cache_lookup_offset(mm_base, poffset, start, &me->multi_list);
-		/* root offset is a special case */
-		if (!oe || !oe->mapent || (strlen(oe->key) - start) == 1)
-			continue;
-
-		strcpy(path, root);
-		strcat(path, poffset);
-		if (umount(path)) {
-			error(ap->logopt, "error recovering from mount fail");
-			error(ap->logopt, "cannot umount offset %s", path);
-		}
-	}
-
-	return;
-}
-
-static int mount_subtree(struct autofs_point *ap, struct mapent *me,
-			 const char *name, char *loc, char *options, void *ctxt)
+static int mount_subtree_offsets(struct autofs_point *ap, struct mapent_cache *mc, struct mapent *me)
 {
 	struct mapent *mm;
-	struct mapent *ro;
-	char *mm_root, *mm_base, *mm_key;
-	const char *mnt_root;
-	unsigned int mm_root_len, mnt_root_len;
-	int start, ret = 0, rv;
-
-	rv = 0;
+	char *m_key;
+	int ret, start;
+	char *base, *m_root;
+	char buf[MAX_ERR_BUF];
 
 	mm = me->multi;
-	mm_key = mm->key;
 
-	if (*mm_key == '/') {
-		mm_root = mm_key;
-		start = strlen(mm_key);
+	if (!mm)
+		return 0;
+
+	cache_multi_lock(me->parent);
+
+	m_key = mm->key;
+
+	if (*m_key == '/') {
+		m_root = m_key;
+		start = strlen(m_key);
 	} else {
-		start = strlen(ap->path) + strlen(mm_key) + 1;
-		mm_root = alloca(start + 3);
-		strcpy(mm_root, ap->path);
-		strcat(mm_root, "/");
-		strcat(mm_root, mm_key);
-	}
-	mm_root_len = strlen(mm_root);
-
-	mnt_root = mm_root;
-	mnt_root_len = mm_root_len;
-
-	if (me == me->multi) {
-		/* name = NULL */
-		/* destination = mm_root */
-		mm_base = "/";
-
-		/* Mount root offset if it exists */
-		ro = cache_lookup_offset(mm_base, mm_base, strlen(mm_root), &me->multi_list);
-		if (ro) {
-			char *myoptions, *ro_loc, *tmp;
-			int namelen = name ? strlen(name) : 0;
-			const char *root;
-			int ro_len;
-
-			rv = parse_mapent(ro->mapent,
-				options, &myoptions, &ro_loc, ap->logopt);
-			if (!rv) {
-				warn(ap->logopt,
-				      MODPREFIX "failed to parse root offset");
-				cache_delete_offset_list(me->mc, name);
-				return 1;
-			}
-			ro_len = 0;
-			if (ro_loc)
-				ro_len = strlen(ro_loc);
-
-			tmp = alloca(mnt_root_len + 2);
-			strcpy(tmp, mnt_root);
-			tmp[mnt_root_len] = '/';
-			tmp[mnt_root_len + 1] = '\0';
-			root = tmp;
-
-			rv = sun_mount(ap, root, name, namelen, ro_loc, ro_len, myoptions, ctxt);
-
-			free(myoptions);
-			if (ro_loc)
-				free(ro_loc);
-		}
-
-		if (ro && rv == 0) {
-			ret = mount_multi_triggers(ap, me, mnt_root, start, mm_base);
-			if (ret == -1) {
-				error(ap->logopt, MODPREFIX
-					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
-				return 1;
-			}
-		} else if (rv <= 0) {
-			ret = mount_multi_triggers(ap, me, mm_root, start, mm_base);
-			if (ret == -1) {
-				error(ap->logopt, MODPREFIX
-					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
-				return 1;
-			}
-		}
-	} else {
-		int loclen = strlen(loc);
-		int namelen = strlen(name);
-
-		mnt_root = name;
-
-		/* name = mm_root + mm_base */
-		/* destination = mm_root + mm_base = name */
-		mm_base = &me->key[start];
-
-		rv = sun_mount(ap, mnt_root, name, namelen, loc, loclen, options, ctxt);
-		if (rv == 0) {
-			ret = mount_multi_triggers(ap, me->multi, mnt_root, start, mm_base);
-			if (ret == -1) {
-				error(ap->logopt, MODPREFIX
-					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
-				return 1;
-			}
-		} else if (rv < 0) {
-			char *mm_root_base = alloca(strlen(mm_root) + strlen(mm_base) + 1);
-	
-			strcpy(mm_root_base, mm_root);
-			strcat(mm_root_base, mm_base);
-
-			ret = mount_multi_triggers(ap, me->multi, mm_root_base, start, mm_base);
-			if (ret == -1) {
-				error(ap->logopt, MODPREFIX
-					 "failed to mount offset triggers");
-				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
-				return 1;
-			}
+		start = strlen(ap->path) + strlen(m_key) + 1;
+		m_root = alloca(start + 1);
+		if (!m_root) {
+			char *estr;
+			cache_multi_unlock(me->parent);
+			estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			error(ap->logopt, MODPREFIX "alloca: %s", estr);
+			return -1;
 		}
+		strcpy(m_root, ap->path);
+		strcat(m_root, "/");
+		strcat(m_root, m_key);
 	}
 
-	/* Mount for base of tree failed */
-	if (rv > 0)
-		return rv;
+	base = &me->key[start];
 
-	/*
-	 * Convert fail on nonstrict, non-empty multi-mount
-	 * to success
-	 */
-	if (rv < 0 && ret > 0)
-		rv = 0;
-
-	return rv;
-}
-
-static char *do_expandsunent(const char *src, const char *key,
-			     const struct substvar *svc, int slashify_colons)
-{
-	char *mapent;
-	int len;
-
-	len = expandsunent(src, NULL, key, svc, slashify_colons);
-	if (len == 0) {
-		errno = EINVAL;
-		return NULL;
+	ret = mount_multi_triggers(ap, m_root, me->multi, base);
+	if (ret == -1) {
+		cache_multi_unlock(me->parent);
+		error(ap->logopt, MODPREFIX "failed to mount offset triggers");
+		return -1;
 	}
-	len++;
 
-	mapent = malloc(len);
-	if (!mapent)
-		return NULL;
-	memset(mapent, 0, len);
+	cache_multi_unlock(me->parent);
 
-	expandsunent(src, mapent, key, svc, slashify_colons);
-
-	return mapent;
+	return ret;
 }
 
 /*
@@ -1276,11 +1023,11 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	char buf[MAX_ERR_BUF];
 	struct map_source *source;
 	struct mapent_cache *mc;
-	struct mapent *me;
+	struct mapent *me, *ro;
 	char *pmapent, *options;
 	const char *p;
 	int mapent_len, rv = 0;
-	int cur_state;
+	int optlen, cur_state;
 	int slashify = ctxt->slashify_colons;
 	unsigned int append_options;
 
@@ -1296,48 +1043,34 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	}
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-
-	/* Offset map entries have been expanded already, avoid expanding
-	 * them again so that the quote handling is consistent between map
-	 * entry locations and (previously expanded) offset map entry
-	 * locations.
-	 */
-	if (*name == '/') {
-		cache_readlock(mc);
-		me = cache_lookup_distinct(mc, name);
-		if (me && me->multi && me->multi != me) {
-			cache_unlock(mc);
-			mapent_len = strlen(mapent) + 1;
-			pmapent = malloc(mapent_len + 1);
-			if (!pmapent) {
-				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-				logerr(MODPREFIX "malloc: %s", estr);
-				return 1;
-			}
-			memset(pmapent, 0, mapent_len + 1);
-			strcpy(pmapent, mapent);
-			goto dont_expand;
-		}
-		cache_unlock(mc);
-	}
-
 	macro_lock();
-	ctxt->subst = addstdenv(ctxt->subst, NULL);
 
-	pmapent = do_expandsunent(mapent, name, ctxt->subst, slashify);
-	if (!pmapent) {
+	ctxt->subst = addstdenv(ctxt->subst);
+
+	mapent_len = expandsunent(mapent, NULL, name, ctxt->subst, slashify);
+	if (mapent_len == 0) {
 		error(ap->logopt, MODPREFIX "failed to expand map entry");
-		ctxt->subst = removestdenv(ctxt->subst, NULL);
+		ctxt->subst = removestdenv(ctxt->subst);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
 	}
-	mapent_len = strlen(pmapent) + 1;
 
-	ctxt->subst = removestdenv(ctxt->subst, NULL);
+	pmapent = alloca(mapent_len + 1);
+	if (!pmapent) {	
+		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+		logerr(MODPREFIX "alloca: %s", estr);
+		ctxt->subst = removestdenv(ctxt->subst);
+		macro_unlock();
+		pthread_setcancelstate(cur_state, NULL);
+		return 1;
+	}
+	pmapent[mapent_len] = '\0';
+
+	expandsunent(mapent, pmapent, name, ctxt->subst, slashify);
+	ctxt->subst = removestdenv(ctxt->subst);
+
 	macro_unlock();
-
-dont_expand:
 	pthread_setcancelstate(cur_state, NULL);
 
 	debug(ap->logopt, MODPREFIX "expanded entry: %s", pmapent);
@@ -1346,10 +1079,10 @@ dont_expand:
 	options = strdup(ctxt->optstr ? ctxt->optstr : "");
 	if (!options) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		free(pmapent);
 		logerr(MODPREFIX "strdup: %s", estr);
 		return 1;
 	}
+	optlen = strlen(options);
 
 	p = skipspace(pmapent);
 
@@ -1375,7 +1108,6 @@ dont_expand:
 					if (mnt_options)
 						free(mnt_options);
 					free(options);
-					free(pmapent);
 					return 1;
 				}
 				mnt_options = tmp;
@@ -1402,7 +1134,6 @@ dont_expand:
 						free(options);
 					if (mnt_options)
 						free(mnt_options);
-					free(pmapent);
 					return 1;
 				}
 				options = tmp;
@@ -1416,8 +1147,8 @@ dont_expand:
 	if (check_is_multi(p)) {
 		char *m_root = NULL;
 		int m_root_len;
-		time_t age;
-		int l;
+		time_t age = time(NULL);
+		int l, ret;
 
 		/* If name starts with "/" it's a direct mount */
 		if (*name == '/') {
@@ -1426,7 +1157,6 @@ dont_expand:
 			if (!m_root) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				free(options);
-				free(pmapent);
 				logerr(MODPREFIX "alloca: %s", estr);
 				return 1;
 			}
@@ -1437,7 +1167,6 @@ dont_expand:
 			if (!m_root) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 				free(options);
-				free(pmapent);
 				logerr(MODPREFIX "alloca: %s", estr);
 				return 1;
 			}
@@ -1446,53 +1175,52 @@ dont_expand:
 			strcat(m_root, name);
 		}
 
-		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		cache_readlock(mc);
+		cache_writelock(mc);
 		me = cache_lookup_distinct(mc, name);
+		if (!me) {
+			int ret;
+			/*
+			 * Not in the cache, perhaps it's a program map
+			 * or one that doesn't support enumeration
+			 */
+			ret = cache_add(mc, source, name, mapent, time(NULL));
+			if (ret == CHE_FAIL) {
+				cache_unlock(mc);
+				free(options);
+				return 1;
+			}
+		} else {
+			/*
+			 * If the entry exists it must not have any existing
+			 * multi-mount subordinate entries since we are
+			 * mounting this afresh. We need to do this to allow
+			 * us to fail on the check for duplicate offsets in
+			 * we don't know when submounts go away.
+			 */
+			cache_multi_lock(me);
+			cache_delete_offset_list(mc, name);
+			cache_multi_unlock(me);
+		}
+		cache_unlock(mc);
+
+		cache_readlock(mc);
+		me = cache_lookup_distinct(mc, name);
+		if (me) {
+			/* So we know we're the multi-mount root */
+			if (!me->multi)
+				me->multi = me;
+		}
+
 		if (!me) {
 			free(options);
-			free(pmapent);
 			cache_unlock(mc);
-			pthread_setcancelstate(cur_state, NULL);
 			error(ap->logopt,
 			      MODPREFIX "can't find multi root %s", name);
 			return 1;
 		}
 
-		cache_multi_writelock(me);
-		/* So we know we're the multi-mount root */
-		if (!me->multi)
-			me->multi = me;
-		else {
-			/*
-			 * The amd host mount type assumes the lookup name
-			 * is the host name for the host mount but amd uses
-			 * ${rhost} for this.
-			 *
-			 * This introduces the possibility of multiple
-			 * concurrent mount requests since constructing a
-			 * mount tree that isn't under the lookup name can't
-			 * take advantage of the kernel queuing of other
-			 * concurrent lookups while the mount tree is
-			 * constructed.
-			 *
-			 * Consequently multi-mount updates (currently only
-			 * done for the internal hosts map which the amd
-			 * parser also uses for its hosts map) can't be
-			 * allowed for amd mounts.
-			 */
-			if (source->flags & MAP_FLAG_FORMAT_AMD) {
-				free(options);
-				free(pmapent);
-				cache_multi_unlock(me);
-				cache_unlock(mc);
-				pthread_setcancelstate(cur_state, NULL);
-				return 0;
-			}
-		}
-
-		age = me->age;
-
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+		cache_multi_lock(me);
 		/* It's a multi-mount; deal with it */
 		do {
 			char *path, *myoptions, *loc;
@@ -1516,7 +1244,6 @@ dont_expand:
 				cache_multi_unlock(me);
 				cache_unlock(mc);
 				free(options);
-				free(pmapent);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
@@ -1531,7 +1258,6 @@ dont_expand:
 				cache_unlock(mc);
 				free(path);
 				free(options);
-				free(pmapent);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
@@ -1542,9 +1268,9 @@ dont_expand:
 			master_source_current_wait(ap->entry);
 			ap->entry->current = source;
 
-			status = update_offset_entry(ap, name,
-						     m_root, m_root_len,
-						     path, myoptions, loc, age);
+			status = add_offset_entry(ap, name,
+						m_root, m_root_len,
+						path, myoptions, loc, age);
 
 			if (status != CHE_OK) {
 				warn(ap->logopt, MODPREFIX "error adding multi-mount");
@@ -1553,36 +1279,70 @@ dont_expand:
 				cache_unlock(mc);
 				free(path);
 				free(options);
-				free(pmapent);
 				free(myoptions);
-				if (loc)
-					free(loc);
+				free(loc);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
 
-			if (loc)
-				free(loc);
+			free(loc);
 			free(path);
 			free(myoptions);
 		} while (*p == '/' || (*p == '"' && *(p + 1) == '/'));
 
 		/*
 		 * We've got the ordered list of multi-mount entries so go
-		 * through and remove any stale entries if this is the top
-		 * of the multi-mount and set the parent entry of each.
+		 * through and set the parent entry of each
 		 */
-		if (me == me->multi)
-			clean_stale_multi_triggers(ap, me, NULL, NULL);
 		cache_set_parents(me);
 
-		rv = mount_subtree(ap, me, name, NULL, options, ctxt);
+		/* Mount root offset if it exists */
+		ro = cache_lookup_offset("/", "/", strlen(m_root), &me->multi_list);
+		if (ro) {
+			char *myoptions, *loc;
+
+			rv = parse_mapent(ro->mapent,
+				options, &myoptions, &loc, ap->logopt);
+			if (!rv) {
+				warn(ap->logopt,
+				      MODPREFIX "failed to mount root offset");
+				cache_delete_offset_list(mc, name);
+				cache_multi_unlock(me);
+				cache_unlock(mc);
+				free(options);
+				pthread_setcancelstate(cur_state, NULL);
+				return 1;
+			}
+
+			rv = sun_mount(ap, m_root,
+				"/", 1, loc, strlen(loc), myoptions, ctxt);
+
+			free(myoptions);
+			free(loc);
+		}
+
+		ret = mount_multi_triggers(ap, m_root, me, "/");
+		if (ret == -1) {
+			warn(ap->logopt,
+			      MODPREFIX "failed to mount offset triggers");
+			cache_multi_unlock(me);
+			cache_unlock(mc);
+			free(options);
+			pthread_setcancelstate(cur_state, NULL);
+			return 1;
+		}
+
+		/*
+		 * Convert fail on nonstrict, non-empty multi-mount
+		 * to success
+		 */
+		if (rv < 0 && ret > 0)
+			rv = 0;
 
 		cache_multi_unlock(me);
 		cache_unlock(mc);
 
 		free(options);
-		free(pmapent);
 		pthread_setcancelstate(cur_state, NULL);
 
 		return rv;
@@ -1597,33 +1357,22 @@ dont_expand:
 		 * it's already been parsed (above) and any option string
 		 * has already been stripped so just use the remainder.
 		 */
-		cache_readlock(mc);
 		if (*name == '/' &&
 		   (me = cache_lookup_distinct(mc, name)) && me->multi) {
 			loc = strdup(p);
 			if (!loc) {
 				free(options);
-				free(pmapent);
-				cache_unlock(mc);
 				warn(ap->logopt, MODPREFIX "out of memory");
 				return 1;
 			}
-			cache_multi_writelock(me);
-			rv = mount_subtree(ap, me, name, loc, options, ctxt);
-			cache_multi_unlock(me);
-			cache_unlock(mc);
-			free(loc);
-			free(options);
-			free(pmapent);
-			return rv;
+			loclen = strlen(p);
+			goto mount_it;
 		}
-		cache_unlock(mc);
 
 		l = chunklen(p, check_colon(p));
 		loc = dequote(p, l, ap->logopt);
 		if (!loc) {
 			free(options);
-			free(pmapent);
 			warn(ap->logopt, MODPREFIX "null location or out of memory");
 			return 1;
 		}
@@ -1631,17 +1380,15 @@ dont_expand:
 		/* Location can't begin with a '/' */
 		if (*p == '/') {
 			free(options);
-			free(pmapent);
-			free(loc);
 			warn(ap->logopt,
 			      MODPREFIX "error location begins with \"/\"");
 			return 1;
 		}
 
-		if (!validate_location(ap->logopt, loc)) {
+		if (!validate_location(loc)) {
+			warn(ap->logopt, MODPREFIX "invalid location %s", loc);
 			free(loc);
 			free(options);
-			free(pmapent);
 			return 1;
 		}
 
@@ -1659,17 +1406,17 @@ dont_expand:
 			if (!ent) {
 				free(loc);
 				free(options);
-				free(pmapent);
 				warn(ap->logopt,
 				     MODPREFIX "null location or out of memory");
 				return 1;
 			}
 
-			if (!validate_location(ap->logopt, ent)) {
+			if (!validate_location(ent)) {
+				warn(ap->logopt,
+				     MODPREFIX "invalid location %s", loc);
 				free(ent);
 				free(loc);
 				free(options);
-				free(pmapent);
 				return 1;
 			}
 
@@ -1681,7 +1428,6 @@ dont_expand:
 				free(ent);
 				free(loc);
 				free(options);
-				free(pmapent);
 				error(ap->logopt, MODPREFIX "out of memory");
 				return 1;
 			}
@@ -1696,44 +1442,38 @@ dont_expand:
 			p = skipspace(p);
 		}
 
-		/*
-		 * If options are asking for a hosts map loc should be
-		 * NULL but we see it can contain junk, so ....
-		 */
-		if ((strstr(options, "fstype=autofs") &&
-		     strstr(options, "hosts"))) {
-			if (loc) {
-				free(loc);
-				loc = NULL;
-			}
-			loclen = 0;
-		} else {
-			loclen = strlen(loc);
-			if (loclen == 0) {
-				free(loc);
-				free(options);
-				free(pmapent);
-				error(ap->logopt,
-				      MODPREFIX "entry %s is empty!", name);
-				return 1;
-			}
+		loclen = strlen(loc);
+		if (loclen == 0) {
+			free(loc);
+			free(options);
+			error(ap->logopt,
+			      MODPREFIX "entry %s is empty!", name);
+			return 1;
 		}
-
+mount_it:
 		debug(ap->logopt,
 		      MODPREFIX "core of entry: options=%s, loc=%.*s",
 		      options, loclen, loc);
 
-		if (!strcmp(ap->path, "/-"))
-			rv = sun_mount(ap, name, name, name_len,
-				       loc, loclen, options, ctxt);
-		else
-			rv = sun_mount(ap, ap->path, name, name_len,
-				       loc, loclen, options, ctxt);
+		rv = sun_mount(ap, ap->path, name, name_len, loc, loclen, options, ctxt);
 
-		if (loc)
-			free(loc);
+		free(loc);
 		free(options);
-		free(pmapent);
+
+		/*
+		 * If it's a multi-mount insert the triggers
+		 * These are always direct mount triggers so root = ""
+		 */
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+		cache_readlock(mc);
+		me = cache_lookup_distinct(mc, name);
+		if (me) {
+			int ret = mount_subtree_offsets(ap, mc, me);
+			/* Convert fail on nonstrict, non-empty multi-mount to success */
+			if (rv < 0 && ret > 0)
+				rv = 0;
+		}
+		cache_unlock(mc);
 		pthread_setcancelstate(cur_state, NULL);
 	}
 	return rv;
@@ -1744,12 +1484,10 @@ int parse_done(void *context)
 	int rv = 0;
 	struct parse_context *ctxt = (struct parse_context *) context;
 
-	instance_mutex_lock();
 	if (--init_ctr == 0) {
 		rv = close_mount(mount_nfs);
 		mount_nfs = NULL;
 	}
-	instance_mutex_unlock();
 	if (ctxt)
 		kill_context(ctxt);
 
